from norch.tensor import Tensor
from .nn import *
from .optim import *
from .utils import *
from .norchvision import *

__version__ = "0.0.4"
__author__ = 'Lucas de Lima Nogueira'
__credits__ = 'Lucas de Lima Nogueira'


import unittest
import norch
from norch.utils import utils_unittests as utils
import torch
import os

class TestTensorAutograd(unittest.TestCase):
    def setUp(self):
        self.device = os.environ.get('device')
        if self.device is None or self.device != 'cuda':
            self.device = 'cpu'
        
        print(f"Running tests on: {self.device}")

        
    def test_addition(self):
        """
        Test autograd from addition two tensors: tensor1 + tensor2
        """
        norch_tensor1 = norch.Tensor([[[1, 2.5], [3, -4]], [[5, 6], [7, 8]]], requires_grad=True).to(self.device)
        norch_tensor2 = norch.Tensor([[[1, 1.], [1, 1.9]], [[1, 1], [1, 1]]], requires_grad=True).to(self.device)
        norch_result = (norch_tensor1 + norch_tensor2).sum()
        norch_result.backward()
        norch_tensor1_grad = utils.to_torch(norch_tensor1.grad).to(self.device)
        norch_tensor2_grad = utils.to_torch(norch_tensor2.grad).to(self.device)

        torch_tensor1 = torch.tensor([[[1, 2.5], [3, -4]], [[5, 6], [7, 8]]], requires_grad=True, device=self.device)
        torch_tensor2 = torch.tensor([[[1, 1.], [1, 1.9]], [[1, 1], [1, 1]]], requires_grad=True, device=self.device)
        torch_result = (torch_tensor1 + torch_tensor2).sum()
        torch_result.backward()
        torch_tensor1_grad = torch_tensor1.grad
        torch_tensor2_grad = torch_tensor2.grad

        self.assertTrue(utils.compare_torch(norch_tensor1_grad, torch_tensor1_grad))
        self.assertTrue(utils.compare_torch(norch_tensor2_grad, torch_tensor2_grad))

    def test_sum_axis(self):
        """
        Test autograd from sum specifying axis
        """
        norch_tensor1 = norch.Tensor([[[1, 2.5], [3, -4]], [[5, 6], [7, 8]]], requires_grad=True).to(self.device)
        norch_tensor2 = norch.Tensor([[[1, 1.], [1, 1.9]], [[1, 1], [1, 1]]], requires_grad=True).to(self.device)
        norch_result = (norch_tensor1 + norch_tensor2).sum(axis=0).sum(axis=0).sum()

        norch_result.backward()
        norch_tensor1_grad = utils.to_torch(norch_tensor1.grad).to(self.device)
        norch_tensor2_grad = utils.to_torch(norch_tensor2.grad).to(self.device)

        torch_tensor1 = torch.tensor([[[1, 2.5], [3, -4]], [[5, 6], [7, 8]]], requires_grad=True, device=self.device)
        torch_tensor2 = torch.tensor([[[1, 1.], [1, 1.9]], [[1, 1], [1, 1]]], requires_grad=True, device=self.device)
        
        torch_result = (torch_tensor1 + torch_tensor2).sum(axis=0).sum(axis=0).sum()
        torch_result.backward()
        torch_tensor1_grad = torch_tensor1.grad
        torch_tensor2_grad = torch_tensor2.grad
        
        self.assertTrue(utils.compare_torch(norch_tensor1_grad, torch_tensor1_grad))
        self.assertTrue(utils.compare_torch(norch_tensor2_grad, torch_tensor2_grad))

        norch_tensor1 = norch.Tensor([[[1, 2.5], [3, -4]], [[5, 6], [7, 8]]], requires_grad=True).to(self.device)
        norch_tensor2 = norch.Tensor([[[1, 1.], [1, 1.9]], [[1, 1], [1, 1]]], requires_grad=True).to(self.device)
        norch_result = (norch_tensor1 + norch_tensor2).sum(axis=1).sum()

        norch_result.backward()
        norch_tensor1_grad = utils.to_torch(norch_tensor1.grad).to(self.device)
        norch_tensor2_grad = utils.to_torch(norch_tensor2.grad).to(self.device)

        torch_tensor1 = torch.tensor([[[1, 2.5], [3, -4]], [[5, 6], [7, 8]]], requires_grad=True, device=self.device)
        torch_tensor2 = torch.tensor([[[1, 1.], [1, 1.9]], [[1, 1], [1, 1]]], requires_grad=True, device=self.device)
        
        torch_result = (torch_tensor1 + torch_tensor2).sum(axis=1).sum()
        torch_result.backward()
        torch_tensor1_grad = torch_tensor1.grad
        torch_tensor2_grad = torch_tensor2.grad
        
        self.assertTrue(utils.compare_torch(norch_tensor1_grad, torch_tensor1_grad))
        self.assertTrue(utils.compare_torch(norch_tensor2_grad, torch_tensor2_grad))
    
    
    def test_max(self):
        """
        Test autograd from max
        """
        norch_tensor = norch.Tensor([[[10, 10], [-4, -4]], [[5., 6], [7, 8]]], requires_grad=True).to(self.device)
        norch_result = norch_tensor.max()

        norch_result.backward()
        norch_tensor_grad = utils.to_torch(norch_tensor.grad).to(self.device)

        torch_tensor = torch.tensor([[[10, 10], [-4, -4]], [[5., 6], [7, 8]]], requires_grad=True, device=self.device)
        
        torch_result = torch_tensor.max()
        torch_result.backward()
        torch_tensor_grad = torch_tensor.grad
        
        self.assertTrue(utils.compare_torch(norch_tensor_grad, torch_tensor_grad))
    
    def test_max_axis(self):
        """
        Test autograd from max specifying axis
        """
        norch_tensor = norch.Tensor([[[10, 10], [-4, -4]], [[5., 6], [7, 8]]], requires_grad=True).to(self.device)
        norch_max_axis = norch_tensor.max(axis=1)
        norch_result = norch_max_axis.sum()

        norch_result.backward()
        norch_tensor_grad = utils.to_torch(norch_tensor.grad).to(self.device)

        torch_tensor = torch.tensor([[[10, 10], [-4, -4]], [[5., 6], [7, 8]]], requires_grad=True, device=self.device)

        torch_max_axis, _ = torch_tensor.max(axis=1)
        torch_result = torch_max_axis.sum()
        torch_result.backward()
        torch_tensor_grad = torch_tensor.grad
        
        self.assertTrue(utils.compare_torch(norch_tensor_grad, torch_tensor_grad))

        norch_tensor = norch.Tensor([[[10, 1], [-4, 0]], [[5., 50], [7, 8]]], requires_grad=True).to(self.device)
        norch_max_axis = norch_tensor.max(axis=2)
        norch_result = norch_max_axis.sum()

        norch_result.backward()
        norch_tensor_grad = utils.to_torch(norch_tensor.grad).to(self.device)

        torch_tensor = torch.tensor([[[10, 1], [-4, 0]], [[5., 50], [7, 8]]], requires_grad=True, device=self.device)

        torch_max_axis, _ = torch_tensor.max(axis=2)
        torch_result = torch_max_axis.sum()
        torch_result.backward()
        torch_tensor_grad = torch_tensor.grad
        self.assertTrue(utils.compare_torch(norch_tensor_grad, torch_tensor_grad))


    ## evaluate case with some repeated values and axis 2

    #def test_max_axis(self):
    #    """
    #    Test autograd from max specifying axis
    #    """
    #
    #    norch_tensor = norch.Tensor([[[10, 10], [-4, 0]], [[5., 50], [7, 8]]], requires_grad=True).to(self.device)
    #    norch_max_axis = norch_tensor.max(axis=2)
    #    norch_result = norch_max_axis.sum()
    #
    #    norch_result.backward()
    #    norch_tensor_grad = utils.to_torch(norch_tensor.grad).to(self.device)
    #
    #    torch_tensor = torch.tensor([[[10, 10], [-4, 0]], [[5., 50], [7, 8]]], requires_grad=True).to(self.device)
    #
    #    torch_max_axis, _ = torch_tensor.max(axis=2)
    #    torch_result = torch_max_axis.sum()
    #    torch_result.backward()
    #    torch_tensor_grad = torch_tensor.grad
    #    self.assertTrue(utils.compare_torch(norch_tensor_grad, torch_tensor_grad))


    #def test_min_axis(self):
    #    """
    #    Test autograd from max specifying axis
    #    """
    #
    #    norch_tensor = norch.Tensor([[[10, 10], [-4, 0]], [[5., 50], [7, 8]]], requires_grad=True).to(self.device)
    #    norch_min_axis = norch_tensor.min(axis=2)
    #    norch_result = norch_min_axis.sum()
    #
    #    norch_result.backward()
    #    norch_tensor_grad = utils.to_torch(norch_tensor.grad).to(self.device)
    #
    #    torch_tensor = torch.tensor([[[10, 10], [-4, 0]], [[5., 50], [7, 8]]], requires_grad=True).to(self.device)
    #
    #    torch_min_axis, _ = torch_tensor.min(axis=2)
    #    torch_result = torch_min_axis.sum()
    #    torch_result.backward()
    #    torch_tensor_grad = torch_tensor.grad
    #    self.assertTrue(utils.compare_torch(norch_tensor_grad, torch_tensor_grad))


    
    def test_min(self):
        """
        Test autograd from min
        """
        norch_tensor = norch.Tensor([[[10, 10], [-4, -4]], [[5., 6], [7, 8]]], requires_grad=True).to(self.device)
        norch_result = norch_tensor.min()

        norch_result.backward()
        norch_tensor_grad = utils.to_torch(norch_tensor.grad).to(self.device)

        torch_tensor = torch.tensor([[[10, 10], [-4, -4]], [[5., 6], [7, 8]]], requires_grad=True, device=self.device)
        
        torch_result = torch_tensor.min()
        torch_result.backward()
        torch_tensor_grad = torch_tensor.grad
        
        self.assertTrue(utils.compare_torch(norch_tensor_grad, torch_tensor_grad))

    def test_min_axis(self):
        """
        Test autograd from min specifying axis
        """
        norch_tensor = norch.Tensor([[[10, 10], [-4, -4]], [[5., 6], [7, 8]]], requires_grad=True).to(self.device)
        norch_min = norch_tensor.min(axis=1)
        norch_result = norch_min.sum()

        norch_result.backward()
        norch_tensor_grad = utils.to_torch(norch_tensor.grad).to(self.device)

        torch_tensor = torch.tensor([[[10, 10], [-4, -4]], [[5., 6], [7, 8]]], requires_grad=True, device=self.device)
        
        torch_min, _ = torch_tensor.min(axis=1)
        torch_result = torch_min.sum()
        torch_result.backward()
        torch_tensor_grad = torch_tensor.grad
        
        self.assertTrue(utils.compare_torch(norch_tensor_grad, torch_tensor_grad))

        norch_tensor = norch.Tensor([[[10, 1], [-4, 0]], [[5., 50], [7, 8]]], requires_grad=True).to(self.device)
        norch_min = norch_tensor.min(axis=2)
        norch_result = norch_min.sum()

        norch_result.backward()
        norch_tensor_grad = utils.to_torch(norch_tensor.grad).to(self.device)

        torch_tensor = torch.tensor([[[10, 1], [-4, 0]], [[5., 50], [7, 8]]], requires_grad=True, device=self.device)
        
        torch_min, _ = torch_tensor.min(axis=2)
        torch_result = torch_min.sum()
        torch_result.backward()
        torch_tensor_grad = torch_tensor.grad
        
        self.assertTrue(utils.compare_torch(norch_tensor_grad, torch_tensor_grad))
    

    def test_broadcasted_addition_autograd(self):
        """
        Test autograd for broadcasting addition: tensor1 + tensor2
        """
        norch_tensor1 = norch.Tensor([[[1., 2, 3], [4, 5, 6]]], requires_grad=True).to(self.device)  # Shape (1, 2, 3)
        norch_tensor2 = norch.Tensor([1.5, -1, 0], requires_grad=True).to(self.device)  # Shape (3)
        norch_result = (norch_tensor1 + norch_tensor2).sum()
        norch_result.backward()
        norch_tensor1_grad = utils.to_torch(norch_tensor1.grad).to(self.device)
        norch_tensor2_grad = utils.to_torch(norch_tensor2.grad).to(self.device)

        torch_tensor1 = torch.tensor([[[1., 2, 3], [4, 5, 6]]], requires_grad=True, device=self.device)  # Shape (1, 2, 3)
        torch_tensor2 = torch.tensor([1.5, -1, 0], requires_grad=True, device=self.device)  # Shape (3)
        torch_result = (torch_tensor1 + torch_tensor2).sum()
        torch_result.backward()
        torch_tensor1_grad = torch_tensor1.grad
        torch_tensor2_grad = torch_tensor2.grad

        self.assertTrue(utils.compare_torch(norch_tensor1_grad, torch_tensor1_grad))
        self.assertTrue(utils.compare_torch(norch_tensor2_grad, torch_tensor2_grad))

        ## reversed order broadcasting
        norch_tensor1 = norch.Tensor([[[1., 2, 3], [4, 5, 6]]], requires_grad=True).to(self.device)  # Shape (1, 2, 3)
        norch_tensor2 = norch.Tensor([1.5, -1, 0], requires_grad=True).to(self.device)  # Shape (3)

        norch_result = (norch_tensor2 + norch_tensor1).sum()
        norch_result.backward()
        norch_tensor1_grad = utils.to_torch(norch_tensor1.grad).to(self.device)
        norch_tensor2_grad = utils.to_torch(norch_tensor2.grad).to(self.device)

        torch_tensor1 = torch.tensor([[[1., 2, 3], [4, 5, 6]]], requires_grad=True, device=self.device)  # Shape (1, 2, 3)
        torch_tensor2 = torch.tensor([1.5, -1, 0], requires_grad=True, device=self.device)  # Shape (3)

        torch_result = (torch_tensor2 + torch_tensor1).sum()
        torch_result.backward()
        torch_tensor1_grad = torch_tensor1.grad
        torch_tensor2_grad = torch_tensor2.grad

        self.assertTrue(utils.compare_torch(norch_tensor1_grad, torch_tensor1_grad))
        self.assertTrue(utils.compare_torch(norch_tensor2_grad, torch_tensor2_grad))
    
    def test_subtraction(self):
        """
        Test autograd from subtraction two tensors: tensor1 - tensor2
        """
        norch_tensor1_sub = norch.Tensor([[[1, 2.5], [3, -4]], [[5, 6], [7, 8]]], requires_grad=True).to(self.device)
        norch_tensor2_sub = norch.Tensor([[[1, 1.], [1, 1.9]], [[1, 1], [1, 1]]], requires_grad=True).to(self.device)
        norch_result_sub = (norch_tensor1_sub - norch_tensor2_sub).sum()
        norch_result_sub.backward()
        norch_tensor1_grad_sub = utils.to_torch(norch_tensor1_sub.grad).to(self.device)
        norch_tensor2_grad_sub = utils.to_torch(norch_tensor2_sub.grad).to(self.device)

        torch_tensor1_sub = torch.tensor([[[1, 2.5], [3, -4]], [[5, 6], [7, 8]]], requires_grad=True, device=self.device)
        torch_tensor2_sub = torch.tensor([[[1, 1.], [1, 1.9]], [[1, 1], [1, 1]]], requires_grad=True, device=self.device)
        torch_result_sub = (torch_tensor1_sub - torch_tensor2_sub).sum()
        torch_result_sub.backward()
        torch_tensor1_grad_sub = torch_tensor1_sub.grad
        torch_tensor2_grad_sub = torch_tensor2_sub.grad

        self.assertTrue(utils.compare_torch(norch_tensor1_grad_sub, torch_tensor1_grad_sub))
        self.assertTrue(utils.compare_torch(norch_tensor2_grad_sub, torch_tensor2_grad_sub))
        
    def test_broadcasted_subtraction_autograd(self):
        """
        Test autograd for broadcasting subtraction: tensor1 - tensor2
        """
        norch_tensor1 = norch.Tensor([[[1., 2, 3], [4, 5, 6]]], requires_grad=True).to(self.device)  # Shape (1, 2, 3)
        norch_tensor2 = norch.Tensor([1.5, -1, 0], requires_grad=True).to(self.device)  # Shape (3)
        norch_result = (norch_tensor1 - norch_tensor2).sum()
        norch_result.backward()
        norch_tensor1_grad = utils.to_torch(norch_tensor1.grad).to(self.device)
        norch_tensor2_grad = utils.to_torch(norch_tensor2.grad).to(self.device)

        torch_tensor1 = torch.tensor([[[1., 2, 3], [4, 5, 6]]], requires_grad=True, device=self.device)  # Shape (1, 2, 3)
        torch_tensor2 = torch.tensor([1.5, -1, 0], requires_grad=True, device=self.device)  # Shape (3)
        torch_result = (torch_tensor1 - torch_tensor2).sum()
        torch_result.backward()
        torch_tensor1_grad = torch_tensor1.grad
        torch_tensor2_grad = torch_tensor2.grad

        self.assertTrue(utils.compare_torch(norch_tensor1_grad, torch_tensor1_grad))
        self.assertTrue(utils.compare_torch(norch_tensor2_grad, torch_tensor2_grad))

        # reversed order broadcasting
        norch_tensor1 = norch.Tensor([[[1., 2, 3], [4, 5, 6]]], requires_grad=True).to(self.device)  # Shape (1, 2, 3)
        norch_tensor2 = norch.Tensor([1.5, -1, 0], requires_grad=True).to(self.device)  # Shape (3)

        norch_result = (norch_tensor2 - norch_tensor1).sum()
        norch_result.backward()
        norch_tensor1_grad = utils.to_torch(norch_tensor1.grad).to(self.device)
        norch_tensor2_grad = utils.to_torch(norch_tensor2.grad).to(self.device)

        torch_tensor1 = torch.tensor([[[1., 2, 3], [4, 5, 6]]], requires_grad=True, device=self.device)  # Shape (1, 2, 3)
        torch_tensor2 = torch.tensor([1.5, -1, 0], requires_grad=True, device=self.device)  # Shape (3)
        
        torch_result = (torch_tensor2 - torch_tensor1).sum()
        torch_result.backward()
        torch_tensor1_grad = torch_tensor1.grad
        torch_tensor2_grad = torch_tensor2.grad

        self.assertTrue(utils.compare_torch(norch_tensor1_grad, torch_tensor1_grad))
        self.assertTrue(utils.compare_torch(norch_tensor2_grad, torch_tensor2_grad))
    

    def test_division(self):
        """
        Test autograd from dividing two tensors: tensor1 / tensor2
        """
        norch_tensor1_div = norch.Tensor([[[2, 5.1], [6, -8]], [[10, 12], [14, 16]]], requires_grad=True).to(self.device)
        norch_tensor2_div = norch.Tensor([[[1, 1], [2, 2.2]], [[3, 3], [4, 4]]], requires_grad=True).to(self.device)
        norch_result_div = (norch_tensor1_div / norch_tensor2_div).sum()
        norch_result_div.backward()
        norch_tensor1_grad_div = utils.to_torch(norch_tensor1_div.grad).to(self.device)
        norch_tensor2_grad_div = utils.to_torch(norch_tensor2_div.grad).to(self.device)

        torch_tensor1_div = torch.tensor([[[2, 5.1], [6, -8]], [[10, 12], [14, 16]]], requires_grad=True, device=self.device)
        torch_tensor2_div = torch.tensor([[[1, 1], [2, 2.2]], [[3, 3], [4, 4]]], requires_grad=True, device=self.device)
        torch_result_div = (torch_tensor1_div / torch_tensor2_div).sum()
        torch_result_div.backward()
        torch_tensor1_grad_div = torch_tensor1_div.grad
        torch_tensor2_grad_div = torch_tensor2_div.grad

        self.assertTrue(utils.compare_torch(norch_tensor1_grad_div, torch_tensor1_grad_div))
        self.assertTrue(utils.compare_torch(norch_tensor2_grad_div, torch_tensor2_grad_div))
    
    
    def test_tensor_division_scalar(self):
        """
        Test autograd from dividing tensor by scalar: tensor / scalar
        """
        norch_tensor_div_scalar = norch.Tensor([[[2, 4.7], [6, 8]], [[10, 12], [14, 16]]], requires_grad=True).to(self.device)
        scalar = 2
        norch_result_div_scalar = (norch_tensor_div_scalar / scalar).sum()
        norch_result_div_scalar.backward()
        norch_tensor_grad_div_scalar = utils.to_torch(norch_tensor_div_scalar.grad).to(self.device)

        torch_tensor_div_scalar = torch.tensor([[[2, 4.7], [6, 8]], [[10, 12], [14, 16]]], requires_grad=True, device=self.device)
        torch_result_div_scalar = (torch_tensor_div_scalar / scalar).sum()
        torch_result_div_scalar.backward()
        torch_tensor_grad_div_scalar = torch_tensor_div_scalar.grad

        self.assertTrue(utils.compare_torch(norch_tensor_grad_div_scalar, torch_tensor_grad_div_scalar))
    
    
    def test_scalar_division_tensor(self):
        """
        Test autograd from dividing scalar by tensor: scalar / tensor
        """
        scalar = 2
        norch_tensor_scalar_div = norch.Tensor([[[1, 2.23], [3, 4]], [[5, 6], [7, 8]]], requires_grad=True).to(self.device)
        norch_result_scalar_div = (scalar / norch_tensor_scalar_div).sum()
        norch_result_scalar_div.backward()
        norch_tensor_grad_scalar_div = utils.to_torch(norch_tensor_scalar_div.grad).to(self.device)

        torch_tensor_scalar_div = torch.tensor([[[1, 2.23], [3, 4]], [[5, 6], [7, 8]]], requires_grad=True, device=self.device)
        torch_result_scalar_div = (scalar / torch_tensor_scalar_div).sum()
        torch_result_scalar_div.backward()
        torch_tensor_grad_scalar_div = torch_tensor_scalar_div.grad

        self.assertTrue(utils.compare_torch(norch_tensor_grad_scalar_div, torch_tensor_grad_scalar_div))
    
    
    def test_power_scalar_tensor(self):
        """
        Test autograd from scalar raised to tensor: scalar ** tensor
        """
        scalar = 2
        norch_tensor_power_st = norch.Tensor([[[2, 3.21], [4, 2.1]], [[6, 7], [8, 9]]], requires_grad=True).to(self.device)
        norch_result_power_st = (scalar ** norch_tensor_power_st).sum()
        norch_result_power_st.backward()
        norch_tensor_grad_power_st = utils.to_torch(norch_tensor_power_st.grad).to(self.device)

        torch_tensor_power_st = torch.tensor([[[2, 3.21], [4, 2.1]], [[6, 7], [8, 9]]], requires_grad=True, device=self.device)
        torch_result_power_st = (scalar ** torch_tensor_power_st).sum()
        torch_result_power_st.backward()
        torch_tensor_grad_power_st = torch_tensor_power_st.grad

        self.assertTrue(utils.compare_torch(norch_tensor_grad_power_st, torch_tensor_grad_power_st))
    
    def test_power_tensor_scalar(self):
        """
        Test autograd from tensor raised to scalar: tensor ** scalar
        """
        scalar = 2
        norch_tensor_power_ts = norch.Tensor([[[2, 3], [4, 2.1]], [[6, 7], [8, 9]]], requires_grad=True).to(self.device)
        norch_result_power_ts = (norch_tensor_power_ts ** scalar).sum()
        norch_result_power_ts.backward()
        norch_tensor_grad_power_ts = utils.to_torch(norch_tensor_power_ts.grad).to(self.device)

        torch_tensor_power_ts = torch.tensor([[[2, 3], [4, 2.1]], [[6, 7], [8, 9]]], requires_grad=True, device=self.device)
        torch_result_power_ts = (torch_tensor_power_ts ** scalar).sum()
        torch_result_power_ts.backward()
        torch_tensor_grad_power_ts = torch_tensor_power_ts.grad

        self.assertTrue(utils.compare_torch(norch_tensor_grad_power_ts, torch_tensor_grad_power_ts))

    def test_matmul(self):
        """
        Test autograd from matrix multiplication: matmul(tensor1, tensor2)
        """
        norch_tensor1_matmul = norch.Tensor([[[1, 2.1], [3, -4]], [[5, 6], [7, 8]]], requires_grad=True).to(self.device)
        norch_tensor2_matmul = norch.Tensor([[[1.1, 3], [4, 5]], [[6, 7], [8, 9]]], requires_grad=True).to(self.device)
        norch_result_matmul = (norch_tensor1_matmul @ norch_tensor2_matmul).sum()
        norch_result_matmul.backward()
        norch_tensor1_grad_matmul = utils.to_torch(norch_tensor1_matmul.grad).to(self.device)
        norch_tensor2_grad_matmul = utils.to_torch(norch_tensor2_matmul.grad).to(self.device)

        torch_tensor1_matmul = torch.tensor([[[1, 2.1], [3, -4]], [[5, 6], [7, 8]]], requires_grad=True, device=self.device)
        torch_tensor2_matmul = torch.tensor([[[1.1, 3], [4, 5]], [[6, 7], [8, 9]]], requires_grad=True, device=self.device)
        torch_result_matmul = (torch_tensor1_matmul @ torch_tensor2_matmul).sum()
        torch_result_matmul.backward()
        torch_tensor1_grad_matmul = torch_tensor1_matmul.grad
        torch_tensor2_grad_matmul = torch_tensor2_matmul.grad

        self.assertTrue(utils.compare_torch(norch_tensor1_grad_matmul, torch_tensor1_grad_matmul))
        self.assertTrue(utils.compare_torch(norch_tensor2_grad_matmul, torch_tensor2_grad_matmul))
    
    def test_batched_matmul(self):
        """
        Test autograd from batched matrix multiplication: BxMxP = BxNxM @ BxMxP
        """
        B = 3  # Batch size

        norch_tensor1_matmul = norch.Tensor([[[1., 2], [3, -4], [5, 6], [7, 8]] for _ in range(B)], requires_grad=True).to(self.device)
        norch_tensor2_matmul = norch.Tensor([[[2., 3, 1, 0, 4], [5, -1, 2, 3, 0]] for _ in range(B)], requires_grad=True).to(self.device)

        norch_result_matmul = norch_tensor1_matmul @ norch_tensor2_matmul
        norch_result_matmul_sum = norch_result_matmul.sum()  # Sum over all elements
        norch_result_matmul_sum.backward()

        # Convert gradients to torch tensors
        norch_tensor1_grad_matmul = utils.to_torch(norch_tensor1_matmul.grad).to(self.device)
        norch_tensor2_grad_matmul = utils.to_torch(norch_tensor2_matmul.grad).to(self.device)

        # Repeat the same process with torch tensors
        torch_tensor1_matmul = torch.tensor([[[1., 2], [3, -4], [5, 6], [7, 8]] for _ in range(B)], requires_grad=True, device=self.device)
        torch_tensor2_matmul = torch.tensor([[[2., 3, 1, 0, 4], [5, -1, 2, 3, 0]] for _ in range(B)], requires_grad=True, device=self.device)
        torch_result_matmul = torch.matmul(torch_tensor1_matmul, torch_tensor2_matmul)
        torch_result_matmul_sum = torch_result_matmul.sum()
        torch_result_matmul_sum.backward()
        
        # Extract gradients from torch tensors
        torch_tensor1_grad_matmul = torch_tensor1_matmul.grad
        torch_tensor2_grad_matmul = torch_tensor2_matmul.grad

        # Assertions to compare the gradients
        self.assertTrue(utils.compare_torch(norch_tensor1_grad_matmul, torch_tensor1_grad_matmul))
        self.assertTrue(utils.compare_torch(norch_tensor2_grad_matmul, torch_tensor2_grad_matmul))

    def test_broadcasted_batched_matmul(self):
        """
        Test autograd from broadcasted batched matrix multiplication: BxMxP = NxM @ BxMxP
        """
        B = 3  # Batch size

        norch_tensor1_matmul = norch.Tensor([[1., 2], [3, -4], [5, 6], [7, 8]], requires_grad=True).to(self.device)
        norch_tensor2_matmul = norch.Tensor([[[2., 3, 1, 0, 4], [5, -1, 2, 3, 0]] for _ in range(B)], requires_grad=True).to(self.device)

        norch_result_matmul = norch_tensor1_matmul @ norch_tensor2_matmul
        norch_result_matmul_sum = norch_result_matmul.sum()  # Sum over all elements
        norch_result_matmul_sum.backward()

        # Convert gradients to torch tensors
        norch_tensor1_grad_matmul = utils.to_torch(norch_tensor1_matmul.grad).to(self.device)
        norch_tensor2_grad_matmul = utils.to_torch(norch_tensor2_matmul.grad).to(self.device)

        # Repeat the same process with torch tensors
        torch_tensor1_matmul = torch.tensor([[1., 2], [3, -4], [5, 6], [7, 8]], requires_grad=True, device=self.device)
        torch_tensor2_matmul = torch.tensor([[[2., 3, 1, 0, 4], [5, -1, 2, 3, 0]] for _ in range(B)], requires_grad=True, device=self.device)
        torch_result_matmul = torch.matmul(torch_tensor1_matmul, torch_tensor2_matmul)
        torch_result_matmul_sum = torch_result_matmul.sum()
        torch_result_matmul_sum.backward()
        
        # Extract gradients from torch tensors
        torch_tensor1_grad_matmul = torch_tensor1_matmul.grad
        torch_tensor2_grad_matmul = torch_tensor2_matmul.grad

        # Assertions to compare the gradients
        self.assertTrue(utils.compare_torch(norch_tensor1_grad_matmul, torch_tensor1_grad_matmul))
        self.assertTrue(utils.compare_torch(norch_tensor2_grad_matmul, torch_tensor2_grad_matmul))


    def test_elementwise_mul_scalar(self):
        """
        Test autograd from elementwise multiplication with scalar: scalar * tensor
        """
        scalar = 2
        norch_tensor_elemwise_mul_scalar = norch.Tensor([[[1.1, 2], [3, -4]], [[5, 6], [7, 8]]], requires_grad=True).to(self.device)
        norch_result_elemwise_mul_scalar = (scalar * norch_tensor_elemwise_mul_scalar).sum()
        norch_result_elemwise_mul_scalar.backward()
        norch_tensor_grad_elemwise_mul_scalar = utils.to_torch(norch_tensor_elemwise_mul_scalar.grad).to(self.device)

        torch_tensor_elemwise_mul_scalar = torch.tensor([[[1.1, 2], [3, -4]], [[5, 6], [7, 8]]], requires_grad=True, device=self.device)
        torch_result_elemwise_mul_scalar = (scalar * torch_tensor_elemwise_mul_scalar).sum()
        torch_result_elemwise_mul_scalar.backward()
        torch_tensor_grad_elemwise_mul_scalar = torch_tensor_elemwise_mul_scalar.grad

        self.assertTrue(utils.compare_torch(norch_tensor_grad_elemwise_mul_scalar, torch_tensor_grad_elemwise_mul_scalar))
    
    
    def test_elementwise_mul_tensor(self):
        """
        Test autograd from elementwise multiplication between two tensors: tensor1 * tensor2
        """
        norch_tensor1_elemwise_mul = norch.Tensor([[[1, 2.1], [3, -4]], [[5, 6], [7, 8]]], requires_grad=True).to(self.device)
        norch_tensor2_elemwise_mul = norch.Tensor([[[1.1, 3], [4, 5]], [[6, 7], [8, 9]]], requires_grad=True).to(self.device)
        norch_result_elemwise_mul = (norch_tensor1_elemwise_mul * norch_tensor2_elemwise_mul).sum()
        norch_result_elemwise_mul.backward()
        norch_tensor1_grad_elemwise_mul = utils.to_torch(norch_tensor1_elemwise_mul.grad).to(self.device)
        norch_tensor2_grad_elemwise_mul = utils.to_torch(norch_tensor2_elemwise_mul.grad).to(self.device)

        torch_tensor1_elemwise_mul = torch.tensor([[[1, 2.1], [3, -4]], [[5, 6], [7, 8]]], requires_grad=True, device=self.device)
        torch_tensor2_elemwise_mul = torch.tensor([[[1.1, 3], [4, 5]], [[6, 7], [8, 9]]], requires_grad=True, device=self.device)
        torch_result_elemwise_mul = (torch_tensor1_elemwise_mul * torch_tensor2_elemwise_mul).sum()
        torch_result_elemwise_mul.backward()
        torch_tensor1_grad_elemwise_mul = torch_tensor1_elemwise_mul.grad
        torch_tensor2_grad_elemwise_mul = torch_tensor2_elemwise_mul.grad

        self.assertTrue(utils.compare_torch(norch_tensor1_grad_elemwise_mul, torch_tensor1_grad_elemwise_mul))
        self.assertTrue(utils.compare_torch(norch_tensor2_grad_elemwise_mul, torch_tensor2_grad_elemwise_mul))

    def test_sin_tensor(self):
        """
        Test autograd from sin operation: sin(tensor)
        """
        norch_sin_tensor = norch.Tensor([[[2, 3.21], [4, 2.1]], [[6, 7], [8, 9]]], requires_grad=True).to(self.device)
        norch_result_sin_tensor = (norch_sin_tensor.sin()).sum()
        norch_result_sin_tensor.backward()
        torch_result_sin_tensor_grad = utils.to_torch(norch_sin_tensor.grad).to(self.device)

        torch_sin_tensor = torch.tensor([[[2, 3.21], [4, 2.1]], [[6, 7], [8, 9]]], requires_grad=True, device=self.device)
        torch_expected_sin_tensor = (torch.sin(torch_sin_tensor)).sum()
        torch_expected_sin_tensor.backward()
        torch_expected_sin_tensor_grad = torch_sin_tensor.grad

        self.assertTrue(utils.compare_torch(torch_result_sin_tensor_grad, torch_expected_sin_tensor_grad))
    
    def test_cos_tensor(self):
        """
        Test autograd from cosine operation: cos(tensor)
        """
        norch_cos_tensor = norch.Tensor([[[2, 3.21], [4, 2.1]], [[6, 7], [8, 9]]], requires_grad=True).to(self.device)
        norch_result_cos_tensor = (norch_cos_tensor.sin()).sum()
        norch_result_cos_tensor.backward()
        torch_result_cos_tensor_grad = utils.to_torch(norch_cos_tensor.grad).to(self.device)

        torch_cos_tensor = torch.tensor([[[2, 3.21], [4, 2.1]], [[6, 7], [8, 9]]], requires_grad=True, device=self.device)
        torch_expected_cos_tensor = (torch.sin(torch_cos_tensor)).sum()
        torch_expected_cos_tensor.backward()
        torch_expected_cos_tensor_grad = torch_cos_tensor.grad

        self.assertTrue(utils.compare_torch(torch_result_cos_tensor_grad, torch_expected_cos_tensor_grad))

    def test_sigmoid(self):
        """
        Test autograd from sigmoid
        """
        norch_tensor = norch.Tensor([[[10, 10], [-4, -4]], [[5., 6], [7, 8]]], requires_grad=True).to(self.device)
        norch_sigmoid = norch.sigmoid(norch_tensor)
        norch_result = norch_sigmoid.sum()

        norch_result.backward()
        norch_tensor_grad = utils.to_torch(norch_tensor.grad).to(self.device)

        torch_tensor = torch.tensor([[[10, 10], [-4, -4]], [[5., 6], [7, 8]]], requires_grad=True, device=self.device)
        torch_sigmoid = torch.sigmoid(torch_tensor)
        torch_result = torch_sigmoid.sum()

        torch_result.backward()
        torch_tensor_grad = torch_tensor.grad

        self.assertTrue(utils.compare_torch(norch_tensor_grad, torch_tensor_grad))

    def test_mse_loss_autograd(self):
        """
        Test the MSELoss with autograd functionality
        """
        loss_fn_norch = norch.nn.MSELoss()
        loss_fn_torch = torch.nn.MSELoss()

        predictions_norch = norch.Tensor([1.1, 2, 3, 4], requires_grad=True).to(self.device)
        labels_norch = norch.Tensor([4, 3, 2.1, 1]).to(self.device)
        loss_norch = loss_fn_norch.forward(predictions_norch, labels_norch)
        loss_norch.backward()  # Backpropagate the loss
        grad_norch = predictions_norch.grad

        predictions_torch = torch.tensor([1.1, 2, 3, 4], requires_grad=True, device=self.device)
        labels_torch = torch.tensor([4, 3, 2.1, 1], device=self.device)
        loss_torch_expected = loss_fn_torch(predictions_torch, labels_torch)
        loss_torch_expected.backward()  # Backpropagate the loss
        grad_torch_expected = predictions_torch.grad

        # Convert norch gradient to torch tensor for comparison
        grad_norch_torch = utils.to_torch(grad_norch).to(self.device)

        self.assertTrue(utils.compare_torch(grad_norch_torch, grad_torch_expected))

    def test_cross_entropy_loss_autograd(self):
        """
        Test the CrossEntropyLoss with autograd functionality
        """
        loss_fn_norch = norch.nn.CrossEntropyLoss()
        loss_fn_torch = torch.nn.CrossEntropyLoss()

        # Test case 1: Single class, single sample
        predictions_norch = norch.Tensor([2.0, 1.0, 0.1], requires_grad=True).to(self.device)
        labels_norch = norch.Tensor([0]).to(self.device)
        loss_norch = loss_fn_norch.forward(predictions_norch, labels_norch)

        loss_norch.backward()  # Backpropagate the loss
        grad_norch = predictions_norch.grad

        predictions_torch = torch.tensor([2.0, 1.0, 0.1], requires_grad=True, device=self.device)
        labels_torch = torch.tensor(0, device=self.device)
        loss_torch_expected = loss_fn_torch(predictions_torch, labels_torch)
        loss_torch_expected.backward()  # Backpropagate the loss
        grad_torch_expected = predictions_torch.grad

        # Convert norch gradient to torch tensor for comparison
        grad_norch_torch = utils.to_torch(grad_norch).to(self.device)

        self.assertTrue(utils.compare_torch(grad_norch_torch, grad_torch_expected))

        # Test case 2: Multiple classes, multiple samples
        predictions_norch = norch.Tensor([[0.5, 1.5, 2.5], [1.0, 2.0, 3.0]], requires_grad=True).to(self.device)
        labels_norch = norch.Tensor([2, 1]).to(self.device)
        loss_norch = loss_fn_norch.forward(predictions_norch, labels_norch)
        loss_norch.backward()  # Backpropagate the loss
        grad_norch = predictions_norch.grad

        predictions_torch = torch.tensor([[0.5, 1.5, 2.5], [1.0, 2.0, 3.0]], requires_grad=True, device=self.device)
        labels_torch = torch.tensor([2, 1], device=self.device)
        loss_torch_expected = loss_fn_torch(predictions_torch, labels_torch)
        loss_torch_expected.backward()  # Backpropagate the loss
        grad_torch_expected = predictions_torch.grad

        # Convert norch gradient to torch tensor for comparison
        grad_norch_torch = utils.to_torch(grad_norch).to(self.device)

        self.assertTrue(utils.compare_torch(grad_norch_torch, grad_torch_expected))

        # Test case 3: Edge case - all predictions are zero
        predictions_norch = norch.Tensor([[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]], requires_grad=True).to(self.device)
        labels_norch = norch.Tensor([1, 2]).to(self.device)
        loss_norch = loss_fn_norch.forward(predictions_norch, labels_norch)
        loss_norch.backward()  # Backpropagate the loss
        grad_norch = predictions_norch.grad

        predictions_torch = torch.tensor([[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]], requires_grad=True, device=self.device)
        labels_torch = torch.tensor([1, 2], device=self.device)
        loss_torch_expected = loss_fn_torch(predictions_torch, labels_torch)
        loss_torch_expected.backward()  # Backpropagate the loss
        grad_torch_expected = predictions_torch.grad

        # Convert norch gradient to torch tensor for comparison
        grad_norch_torch = utils.to_torch(grad_norch).to(self.device)

        self.assertTrue(utils.compare_torch(grad_norch_torch, grad_torch_expected))

        # Test case 4: Batched class probabilities instead of class index
        predictions_norch = norch.Tensor([[0.5, 0.2, 0.1], [0.1, 0.5, 0.7]], requires_grad=True).to(self.device)
        labels_norch = norch.Tensor([[1., 0, 0], [0, 1, 0]]).to(self.device)
        loss_norch = loss_fn_norch.forward(predictions_norch, labels_norch)
        loss_norch.backward()  # Backpropagate the loss
        grad_norch = predictions_norch.grad

        predictions_torch = torch.tensor([[0.5, 0.2, 0.1], [0.1, 0.5, 0.7]], requires_grad=True, device=self.device)
        labels_torch = torch.tensor([[1., 0, 0], [0, 1, 0]], device=self.device)
        loss_torch_expected = loss_fn_torch(predictions_torch, labels_torch)
        loss_torch_expected.backward()  # Backpropagate the loss
        grad_torch_expected = predictions_torch.grad

        # Convert norch gradient to torch tensor for comparison
        grad_norch_torch = utils.to_torch(grad_norch).to(self.device)

        self.assertTrue(utils.compare_torch(grad_norch_torch, grad_torch_expected))


    # implement grad pure softmax --> 0
    # def test_softmax(self):
    #     """
    #     Test autograd from softmax
    #     """
    #     norch_tensor = norch.Tensor([[[-5, 10], [-4, -4]], [[5., 6], [7, 8]]], requires_grad=True).to(self.device)
    #     norch_softmax = norch.softmax(norch_tensor, dim=1)
    #     norch_result = norch_softmax.sum()

    #     norch_result.backward()
    #     norch_tensor_grad = utils.to_torch(norch_tensor.grad).to(self.device)

    #     torch_tensor = torch.tensor([[[-5, 10], [-4, -4]], [[5., 6], [7, 8]]], requires_grad=True).to(self.device)
    #     torch_softmax = torch.softmax(torch_tensor, dim=1)
    #     torch_result = torch_softmax.sum()

    #     torch_result.backward()
    #     torch_tensor_grad = torch_tensor.grad
        
    #     self.assertTrue(utils.compare_torch(norch_tensor_grad, torch_tensor_grad))

    #     norch_tensor = norch.Tensor([[[10, 1], [-4, 0]], [[5., 50], [7, 8]]], requires_grad=True).to(self.device)
    #     norch_softmax = norch.softmax(norch_tensor, dim=2)
    #     norch_result = norch_softmax.sum()

    #     norch_result.backward()
    #     norch_tensor_grad = utils.to_torch(norch_tensor.grad).to(self.device)

    #     torch_tensor = torch.tensor([[[10, 1], [-4, 0]], [[5., 50], [7, 8]]], requires_grad=True).to(self.device)

    #     torch_softmax = torch.softmax(torch_tensor, dim=2)
    #     torch_result = torch_softmax.sum()
    #     torch_result.backward()
    #     torch_tensor_grad = torch_tensor.grad
       
    #     self.assertTrue(utils.compare_torch(norch_tensor_grad, torch_tensor_grad))

    
    def test_reshape(self):
        """
        Test autograd from reshaping a tensor: tensor.reshape(shape)
        """
        norch_tensor_reshape = norch.Tensor([[[1, 2.1], [3, -4]], [[5, 6], [7, 8]]], requires_grad=True).to(self.device)
        new_shape = [2, 4]
        norch_result_reshape = norch_tensor_reshape.reshape(new_shape).sum()
        norch_result_reshape.backward()
        norch_tensor_grad_reshape = utils.to_torch(norch_tensor_reshape.grad).to(self.device)

        torch_tensor_reshape = torch.tensor([[[1, 2.1], [3, -4]], [[5, 6], [7, 8]]], requires_grad=True, device=self.device)
        torch_result_reshape = torch_tensor_reshape.reshape(new_shape).sum()
        torch_result_reshape.backward()
        torch_tensor_grad_reshape = torch_tensor_reshape.grad

        self.assertTrue(utils.compare_torch(norch_tensor_grad_reshape, torch_tensor_grad_reshape))
    
    
    def test_transpose_axes(self):
        """
        Test autograd from transposing a tensor with specific axes: tensor.transpose(axis1, axis2)
        """
        norch_tensor_transpose = norch.Tensor([[[1, 2.1], [3, -4]], [[5, 6], [7, 8]]], requires_grad=True).to(self.device)
        axis1, axis2 = 0, 2
        norch_result_transpose = norch_tensor_transpose.transpose(axis1, axis2).sum()
        norch_result_transpose.backward()
        norch_tensor_grad_transpose = utils.to_torch(norch_tensor_transpose.grad).to(self.device)

        torch_tensor_transpose = torch.tensor([[[1, 2.1], [3, -4]], [[5, 6], [7, 8]]], requires_grad=True, device=self.device)
        torch_result_transpose = torch_tensor_transpose.transpose(axis1, axis2).sum()
        torch_result_transpose.backward()
        torch_tensor_grad_transpose = torch_tensor_transpose.grad

        self.assertTrue(utils.compare_torch(norch_tensor_grad_transpose, torch_tensor_grad_transpose))
    
    
    def test_T(self):
        """
        Test autograd from transposing a tensor using .T attribute
        """
        norch_tensor_T = norch.Tensor([[[1, 2.1], [3, -4]], [[5, 6], [7, 8]]], requires_grad=True).to(self.device)
        norch_result_T = norch_tensor_T.T.sum()
        norch_result_T.backward()
        norch_tensor_grad_T = utils.to_torch(norch_tensor_T.grad).to(self.device)

        torch_tensor_T = torch.tensor([[[1, 2.1], [3, -4]], [[5, 6], [7, 8]]], requires_grad=True, device=self.device)
        torch_result_T = torch_tensor_T.mT.sum()
        torch_result_T.backward()
        torch_tensor_grad_T = torch_tensor_T.grad

        self.assertTrue(utils.compare_torch(norch_tensor_grad_T, torch_tensor_grad_T))

    def test_reshape_then_matmul(self):
        """
        Test autograd from reshaping a tensor then performing matrix multiplication: matmul(tensor1.reshape(shape), tensor2)
        """
        norch_tensor1 = norch.Tensor([[1, 2.1], [3, -4], [5, 6], [7, 8]], requires_grad=True).to(self.device)
        norch_tensor2 = norch.Tensor([[1, 5.1], [0.1, -4], [0, 6], [7, 8]], requires_grad=True).to(self.device)

        new_shape = [2, 4]

        norch_result_reshape_matmul = (norch_tensor1.reshape(new_shape) @ norch_tensor2).sum()
        norch_result_reshape_matmul.backward()
        norch_tensor_grad_reshape_matmul1 = utils.to_torch(norch_tensor1.grad).to(self.device)
        norch_tensor_grad_reshape_matmul2 = utils.to_torch(norch_tensor2.grad).to(self.device)

        torch_tensor1 = torch.tensor([[1, 2.1], [3, -4], [5, 6], [7, 8]], dtype=torch.float32, requires_grad=True, device=self.device)
        torch_tensor2 = torch.tensor([[1, 5.1], [0.1, -4], [0, 6], [7, 8]], dtype=torch.float32, requires_grad=True, device=self.device)     
        
        torch_result_reshape_matmul = (torch_tensor1.reshape(new_shape) @ torch_tensor2).sum()
        torch_result_reshape_matmul.backward()
        torch_tensor_grad_reshape_matmul1 = torch_tensor1.grad
        torch_tensor_grad_reshape_matmul2 = torch_tensor2.grad
        
        self.assertTrue(utils.compare_torch(norch_tensor_grad_reshape_matmul1, torch_tensor_grad_reshape_matmul1))
        self.assertTrue(utils.compare_torch(norch_tensor_grad_reshape_matmul2, torch_tensor_grad_reshape_matmul2))

    def test_unsqueeze(self):
        """
        Test autograd from unsqueezing a tensor: tensor.unsqueeze(dim)
        """
        
        # Unsqueeze at dim=0
        norch_tensor_unsqueeze = norch.Tensor([[1., 2], [3, 4]], requires_grad=True).to(self.device)
        norch_result_unsqueeze_0 = norch_tensor_unsqueeze.unsqueeze(0).sum()
        norch_result_unsqueeze_0.backward()
        norch_tensor_grad_unsqueeze_0 = utils.to_torch(norch_tensor_unsqueeze.grad).to(self.device)

        torch_tensor_unsqueeze = torch.tensor([[1., 2], [3, 4]], requires_grad=True, device=self.device)
        torch_result_unsqueeze_0 = torch_tensor_unsqueeze.unsqueeze(0).sum()
        torch_result_unsqueeze_0.backward()
        torch_tensor_grad_unsqueeze_0 = torch_tensor_unsqueeze.grad

        self.assertTrue(utils.compare_torch(norch_tensor_grad_unsqueeze_0, torch_tensor_grad_unsqueeze_0))

        # Unsqueeze at dim=1
        norch_tensor_unsqueeze = norch.Tensor([[1., 2.], [3, 4]], requires_grad=True).to(self.device)
        norch_result_unsqueeze_1 = norch_tensor_unsqueeze.unsqueeze(1).sum()
        norch_result_unsqueeze_1.backward()
        norch_tensor_grad_unsqueeze_1 = utils.to_torch(norch_tensor_unsqueeze.grad).to(self.device)

        torch_tensor_unsqueeze = torch.tensor([[1., 2.], [3, 4]], requires_grad=True, device=self.device)
        torch_result_unsqueeze_1 = torch_tensor_unsqueeze.unsqueeze(1).sum()
        torch_result_unsqueeze_1.backward()
        torch_tensor_grad_unsqueeze_1 = torch_tensor_unsqueeze.grad

        self.assertTrue(utils.compare_torch(norch_tensor_grad_unsqueeze_1, torch_tensor_grad_unsqueeze_1))

        # Unsqueeze at dim=2
        norch_tensor_unsqueeze = norch.Tensor([[1., 2], [3, 4]], requires_grad=True).to(self.device)
        norch_result_unsqueeze_2 = norch_tensor_unsqueeze.unsqueeze(2).sum()
        norch_result_unsqueeze_2.backward()
        norch_tensor_grad_unsqueeze_2 = utils.to_torch(norch_tensor_unsqueeze.grad).to(self.device)

        torch_tensor_unsqueeze = torch.tensor([[1., 2], [3, 4]], requires_grad=True, device=self.device)
        torch_result_unsqueeze_2 = torch_tensor_unsqueeze.unsqueeze(2).sum()
        torch_result_unsqueeze_2.backward()
        torch_tensor_grad_unsqueeze_2 = torch_tensor_unsqueeze.grad

        self.assertTrue(utils.compare_torch(norch_tensor_grad_unsqueeze_2, torch_tensor_grad_unsqueeze_2))

    def test_unsqueeze_then_matmul(self):
        """
        Test autograd from unsqueezing a tensor then performing matrix multiplication: matmul(tensor1.unsqueeze(dim), tensor2)
        """
        norch_tensor1 = norch.Tensor([[1, 2], [3, 4]], requires_grad=True).to(self.device)
        norch_tensor2 = norch.Tensor([[1, 2], [3, 4]], requires_grad=True).to(self.device)
        
        # Unsqueeze at dim=0 then matmul
        norch_result_unsqueeze_matmul = (norch_tensor1 @ norch_tensor2.unsqueeze(0)).sum()
        norch_result_unsqueeze_matmul.backward()
        norch_tensor_grad_unsqueeze_matmul1 = utils.to_torch(norch_tensor1.grad).to(self.device)
        norch_tensor_grad_unsqueeze_matmul2 = utils.to_torch(norch_tensor2.grad).to(self.device)

        torch_tensor1 = torch.tensor([[1, 2], [3, 4]], dtype=torch.float32, requires_grad=True, device=self.device)
        torch_tensor2 = torch.tensor([[1, 2], [3, 4]], dtype=torch.float32, requires_grad=True, device=self.device)
        
        torch_result_unsqueeze_matmul = (torch_tensor1 @ torch_tensor2.unsqueeze(0)).sum()
        torch_result_unsqueeze_matmul.backward()
        torch_tensor_grad_unsqueeze_matmul1 = torch_tensor1.grad
        torch_tensor_grad_unsqueeze_matmul2 = torch_tensor2.grad
        
        self.assertTrue(utils.compare_torch(norch_tensor_grad_unsqueeze_matmul1, torch_tensor_grad_unsqueeze_matmul1))
        self.assertTrue(utils.compare_torch(norch_tensor_grad_unsqueeze_matmul2, torch_tensor_grad_unsqueeze_matmul2))

    def test_squeeze(self):
        """
        Test autograd from squeezing a tensor: tensor.squeeze(dim)
        """
        # Squeeze at dim=0
        norch_tensor_squeeze = norch.Tensor([[[1., 2], [3, 4]]], requires_grad=True).to(self.device)
        norch_result_squeeze_0 = norch_tensor_squeeze.squeeze(0).sum()
        norch_result_squeeze_0.backward()
        norch_tensor_grad_squeeze_0 = utils.to_torch(norch_tensor_squeeze.grad).to(self.device)

        torch_tensor_squeeze = torch.tensor([[[1., 2], [3, 4]]], requires_grad=True, device=self.device)
        torch_result_squeeze_0 = torch_tensor_squeeze.squeeze(0).sum()
        torch_result_squeeze_0.backward()
        torch_tensor_grad_squeeze_0 = torch_tensor_squeeze.grad

        self.assertTrue(utils.compare_torch(norch_tensor_grad_squeeze_0, torch_tensor_grad_squeeze_0))

    def test_squeeze_then_matmul(self):
        """
        Test autograd from squeezing a tensor then performing matrix multiplication: matmul(tensor1.squeeze(dim), tensor2)
        """
        norch_tensor1 = norch.Tensor([[[1., 2], [3, 4]]], requires_grad=True).to(self.device)
        norch_tensor2 = norch.Tensor([[[1., 2], [3, 4]]], requires_grad=True).to(self.device)
        
        # Squeeze at dim=0 then matmul
        norch_result_squeeze_matmul = (norch_tensor1.squeeze(0) @ norch_tensor2).sum()
        norch_result_squeeze_matmul.backward()
        norch_tensor_grad_squeeze_matmul1 = utils.to_torch(norch_tensor1.grad,).to(self.device)
        norch_tensor_grad_squeeze_matmul2 = utils.to_torch(norch_tensor2.grad).to(self.device)

        torch_tensor1 = torch.tensor([[[1., 2], [3, 4]]], dtype=torch.float32, requires_grad=True, device=self.device)
        torch_tensor2 = torch.tensor([[[1., 2], [3, 4]]], dtype=torch.float32, requires_grad=True, device=self.device)
        
        torch_result_squeeze_matmul = (torch_tensor1.squeeze(0) @ torch_tensor2).sum()
        torch_result_squeeze_matmul.backward()
        torch_tensor_grad_squeeze_matmul1 = torch_tensor1.grad
        torch_tensor_grad_squeeze_matmul2 = torch_tensor2.grad
        
        self.assertTrue(utils.compare_torch(norch_tensor_grad_squeeze_matmul1, torch_tensor_grad_squeeze_matmul1))
        self.assertTrue(utils.compare_torch(norch_tensor_grad_squeeze_matmul2, torch_tensor_grad_squeeze_matmul2))


    def test_T_then_matmul(self):
        """
        Test autograd from transposing a tensor then performing matrix multiplication: matmul(tensor.T, tensor)
        """
        norch_tensor1 = norch.Tensor([[1, 2.1], [3, -4], [5, 6], [7, 8]], requires_grad=True)
        norch_tensor2 = norch.Tensor([[1, 5.1], [0.1, -4], [0, 6], [7, 8]], requires_grad=True)

        norch_result_T_matmul = (norch_tensor1.T @ norch_tensor2).sum()
        norch_result_T_matmul.backward()
        norch_tensor_grad_T_matmul1 = utils.to_torch(norch_tensor1.grad).to(self.device)
        norch_tensor_grad_T_matmult2 = utils.to_torch(norch_tensor2.grad).to(self.device)

        torch_tensor1 = torch.tensor([[1, 2.1], [3, -4], [5, 6], [7, 8]], dtype=torch.float32, requires_grad=True, device=self.device)
        torch_tensor2 = torch.tensor([[1, 5.1], [0.1, -4], [0, 6], [7, 8]], dtype=torch.float32, requires_grad=True, device=self.device)     
        
        torch_result_T_matmul = (torch_tensor1.T @ torch_tensor2).sum()
        torch_result_T_matmul.backward()
        torch_tensor_grad_T_matmul1 = torch_tensor1.grad
        torch_tensor_grad_T_matmul2 = torch_tensor2.grad
        
        self.assertTrue(utils.compare_torch(norch_tensor_grad_T_matmul1, torch_tensor_grad_T_matmul1))
        self.assertTrue(utils.compare_torch(norch_tensor_grad_T_matmult2, torch_tensor_grad_T_matmul2))

    def todo(self):
        """
        The code has a problem on the following operation
        tensor1.reshape(..) @ tensor1
        print(tensor1.grad)
        (also transpsoe and .T)
        """
        pass

    def test_transpose_axes_then_matmul(self):
        """
        Test autograd from transposing a tensor with specific axes then performing matrix multiplication: matmul(tensor.transpose(axis1, axis2), tensor)
        """
        norch_tensor1 = norch.Tensor([[1, 2.1], [3, -4], [5, 6], [7, 8]], requires_grad=True).to(self.device)
        norch_tensor2 = norch.Tensor([[1, 5.1], [0.1, -4], [0, 6], [7, 8]], requires_grad=True).to(self.device)

        norch_result_transpose_matmul = (norch_tensor1.transpose(0, 1) @ norch_tensor2).sum()
        norch_result_transpose_matmul.backward()
        norch_tensor_grad_transpose_matmul1 = utils.to_torch(norch_tensor1.grad).to(self.device)
        norch_tensor_grad_transpose_matmult2 = utils.to_torch(norch_tensor2.grad).to(self.device)

        torch_tensor1 = torch.tensor([[1, 2.1], [3, -4], [5, 6], [7, 8]], dtype=torch.float32, requires_grad=True, device=self.device)
        torch_tensor2 = torch.tensor([[1, 5.1], [0.1, -4], [0, 6], [7, 8]], dtype=torch.float32, requires_grad=True, device=self.device)  
    
        torch_result_transpose_matmul = (torch_tensor1.T @ torch_tensor2).sum()
        torch_result_transpose_matmul.backward()
        torch_tensor_grad_transpose_matmul1 = torch_tensor1.grad
        torch_tensor_grad_transpose_matmul2 = torch_tensor2.grad
        
        self.assertTrue(utils.compare_torch(norch_tensor_grad_transpose_matmul1, torch_tensor_grad_transpose_matmul1))
        self.assertTrue(utils.compare_torch(norch_tensor_grad_transpose_matmult2, torch_tensor_grad_transpose_matmul2))

if __name__ == '__main__':
    unittest.main()


import torch

def to_torch(custom_tensor):
    shape = custom_tensor.shape
    pytorch_tensor = torch.zeros(shape)
    
    def _iterate_indices(shape):
        if len(shape) == 0:
            yield ()
        else:
            for index in range(shape[0]):
                for sub_indices in _iterate_indices(shape[1:]):
                    yield (index,) + sub_indices
    
    # Iterate over all elements using the custom tensor's __getitem__ method
    for indices in _iterate_indices(shape):
        value = custom_tensor[indices]
        pytorch_tensor[tuple(indices)] = value
    
    return pytorch_tensor

def compare_torch(tensor1, tensor2, epsilon=1e-5):
    diff = torch.abs(tensor1 - tensor2)    
    return torch.all(diff < epsilon)


from abc import ABC
from norch.tensor import Tensor

class Optimizer(ABC):
    """
    Abstract class for optimizers
    """

    def __init__(self, parameters):
        if isinstance(parameters, Tensor):
            raise TypeError("parameters should be an iterable but got {}".format(type(parameters)))
        elif isinstance(parameters, dict):
            parameters = parameters.values()

        self.parameters = list(parameters)

    def step(self):
        raise NotImplementedError
    
    def zero_grad(self):
        for module, name, parameter in self.parameters:
            parameter.zero_grad()


from .module import Module
from norch.autograd.functions import *
import norch
from abc import ABC

class Loss(Module, ABC):
    "Abstract class for loss functions"

    def __init__(self):
        super().__init__()

    def forward(self, predictions, labels):
        raise NotImplementedError
    
    def __call__(self, *inputs):
        return self.forward(*inputs)
    

class MSELoss(Loss):
    def __init__(self):
        super().__init__()

    def forward(self, predictions, target):
        assert target.shape == predictions.shape, \
            "Labels and predictions shape does not match: {} and {}".format(target.shape, predictions.shape)
        
        cost = ((predictions - target) ** 2).sum() / predictions.numel
        return cost
    

class CrossEntropyLoss(Loss):
    def __init__(self):
        super().__init__()

    def forward(self, input, target):
        assert isinstance(input, norch.Tensor), \
            "Cross entropy argument 'input' must be Tensor, not {}".format(type(input))
        
        assert isinstance(target, norch.Tensor), \
            "Cross entropy argument 'target' must be Tensor, not {}".format(type(target))
        if input.ndim > 2:
            input = input.squeeze(-1)

        if input.ndim == 1:
            if target.numel == 1:
                num_classes = input.shape[0]
                target = norch.one_hot_encode(target, num_classes).to(target.device)
                
                logits = norch.softmax(input, dim=0)
                target = target.reshape(logits.shape)
                cost = -(logits.log() * target).sum()
                
            else:
                # target -> class probabilities (one-hot encoded)
                assert target.shape == input.shape, \
                    "Input and target shape does not match: {} and {}".format(input.shape, target.shape)
                logits = norch.softmax(input, dim=0)
                target = target.reshape(logits.shape)
                cost = -(logits.log() * target).sum()


        elif input.ndim == 2:
            if target.ndim > 1:
                target = target.squeeze(-1)
            # batched 
            if target.ndim == 1:
                # target -> Ground truth class indices:
                num_classes = input.shape[1]

                target = norch.one_hot_encode(target, num_classes).to(target.device)
                
                batch_size = input.shape[0]
                logits = norch.softmax(input, dim=1)
                target = target.reshape(logits.shape)
                cost = -(logits.log() * target).sum() / batch_size

            else:
                # target -> class probabilities (one-hot encoded)
                assert target.shape == input.shape, \
                    "Input and target shape does not match: {} and {}".format(input.shape, target.shape)
                
                batch_size = input.shape[0]
                logits = norch.softmax(input, dim=1)
                target = target.reshape(logits.shape)
                cost = -(logits.log() * target).sum() / batch_size

        if input.requires_grad:
            cost.grad_fn = CrossEntropyLossBackward(input, target)

        return cost
            





import ctypes
import os
from .autograd.functions import *

class CTensor(ctypes.Structure):
    _fields_ = [
        ('data', ctypes.POINTER(ctypes.c_float)),
        ('strides', ctypes.POINTER(ctypes.c_int)),
        ('shape', ctypes.POINTER(ctypes.c_int)),
        ('ndim', ctypes.c_int),
        ('size', ctypes.c_int),
        ('device', ctypes.c_char_p)
    ]

class Tensor:
    module_dir = os.path.dirname(os.path.abspath(__file__))
    _C = ctypes.CDLL(os.path.join(module_dir, "libtensor.so"))

    def __init__(self, data=None, device="cpu", requires_grad=False):

        if data != None:
            if isinstance(data, (float, int)):
                data = [data]

            data, shape = self.flatten(data)
            
            self.shape = shape.copy()
            
            self.data_ctype = (ctypes.c_float * len(data))(*data.copy())
            self.shape_ctype = (ctypes.c_int * len(shape))(*shape.copy())
            self.ndim_ctype = ctypes.c_int(len(shape))
            self.device_ctype = device.encode('utf-8')

            self.ndim = len(shape)
            self.device = device

            self.numel = 1
            for s in self.shape:
                self.numel *= s

            self.requires_grad = requires_grad
            self.grad = None
            self.grad_fn = None

            Tensor._C.create_tensor.argtypes = [ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_int), ctypes.c_int, ctypes.c_char_p]
            Tensor._C.create_tensor.restype = ctypes.POINTER(CTensor)
            

            self.tensor = Tensor._C.create_tensor(
                self.data_ctype,
                self.shape_ctype,
                self.ndim_ctype,
                self.device_ctype
            )
        
        else:
            self.tensor = None,
            self.shape = None,
            self.ndim = None,
            self.device = device
            self.requires_grad = None
            self.grad = None
            self.grad_fn = None

    def flatten(self, nested_list):
        def flatten_recursively(nested_list):
            flat_data = []
            shape = []
            if isinstance(nested_list, list):
                for sublist in nested_list:
                    inner_data, inner_shape = flatten_recursively(sublist)
                    flat_data.extend(inner_data)
                shape.append(len(nested_list))
                shape.extend(inner_shape)
            else:
                flat_data.append(nested_list)
            return flat_data, shape
        
        flat_data, shape = flatten_recursively(nested_list)
        return flat_data, shape
    
    def ones_like(self):
        
        Tensor._C.ones_like_tensor.argtypes = [ctypes.POINTER(CTensor)]
        Tensor._C.ones_like_tensor.restype = ctypes.POINTER(CTensor)
        Tensor._C.ones_like_tensor(self.tensor)   

        result_tensor_ptr = Tensor._C.ones_like_tensor(self.tensor)

        result_data = Tensor()
        result_data.tensor = result_tensor_ptr
        result_data.shape = self.shape.copy()
        result_data.ndim = self.ndim
        result_data.device = self.device
        result_data.numel = self.numel
        
        return result_data
    
    def zeros_like(self):
        
        Tensor._C.zeros_like_tensor.argtypes = [ctypes.POINTER(CTensor)]
        Tensor._C.zeros_like_tensor.restype = ctypes.POINTER(CTensor)
        Tensor._C.zeros_like_tensor(self.tensor)   

        result_tensor_ptr = Tensor._C.zeros_like_tensor(self.tensor)

        result_data = Tensor()
        result_data.tensor = result_tensor_ptr
        result_data.shape = self.shape.copy()
        result_data.ndim = self.ndim
        result_data.device = self.device
        result_data.numel = self.numel

        return result_data
    
    def reshape(self, new_shape):
        # Calculate the total number of elements in the tensor
        total_elements = self.numel
        
        # Check for the presence of -1 in new_shape
        if new_shape.count(-1) > 1:
            raise ValueError("Only one dimension can be inferred (set to -1).")
        
        inferred_dim = None
        known_dims_product = 1
        for dim in new_shape:
            if dim == -1:
                inferred_dim = dim
            else:
                known_dims_product *= dim
        
        # Calculate the inferred dimension if -1 is present
        if inferred_dim == -1:
            inferred_dim_size = total_elements // known_dims_product
            new_shape = [inferred_dim_size if dim == -1 else dim for dim in new_shape]

        new_shape_ctype = (ctypes.c_int * len(new_shape))(*new_shape)
        new_ndim_ctype = ctypes.c_int(len(new_shape))
        
        Tensor._C.reshape_tensor.argtypes = [ctypes.POINTER(CTensor), ctypes.POINTER(ctypes.c_int), ctypes.c_int]
        Tensor._C.reshape_tensor.restype = ctypes.POINTER(CTensor)
        result_tensor_ptr = Tensor._C.reshape_tensor(self.tensor, new_shape_ctype, new_ndim_ctype)   

        result_data = Tensor()
        result_data.tensor = result_tensor_ptr
        result_data.shape = new_shape.copy()
        result_data.ndim = len(new_shape)
        result_data.device = self.device
        result_data.numel = self.numel

        result_data.requires_grad = self.requires_grad
        if result_data.requires_grad:
            result_data.grad_fn = ReshapeBackward(self)

        return result_data

    def unsqueeze(self, dim):
        if dim < 0:
            dim = self.ndim + dim + 1

        # Ensure the dimension is valid
        if dim > self.ndim:
            raise ValueError("Dimension out of range (expected to be in range of [0, {0}], but got {1})".format(self.ndim, dim))
        
        # Create the new shape with an extra dimension of size 1
        new_shape = self.shape[:dim] + [1] + self.shape[dim:]
        
        return self.reshape(new_shape)

    def squeeze(self, dim=None):
        if dim is not None:
            if dim < 0:
                dim = self.ndim + dim
            
            # Ensure the dimension is valid
            if dim >= self.ndim or dim < 0:
                raise ValueError("Dimension out of range (expected to be in range of [0, {0}), but got {1})".format(self.ndim, dim))
            
            # Only squeeze the specified dimension if its size is 1
            if self.shape[dim] != 1:
                return self
                #raise ValueError("Dimension {0} does not have size 1 and cannot be squeezed".format(dim))
            
            # Create the new shape without the specified dimension
            new_shape = self.shape[:dim] + self.shape[dim+1:]
        else:
            # Create the new shape by removing all dimensions of size 1
            new_shape = [s for s in self.shape if s != 1]
        
        return self.reshape(new_shape)

    def to(self, device):
        self.device = device
        self.device_ctype = self.device.encode('utf-8')

        Tensor._C.to_device.argtypes = [ctypes.POINTER(CTensor), ctypes.c_char_p]
        Tensor._C.to_device.restype = None
        Tensor._C.to_device(self.tensor, self.device_ctype)

        return self
    
    def backward(self, gradient=None):
        if not self.requires_grad:
            return
                
        if gradient is None:
            if self.shape == [1]:
                gradient = Tensor([1]).to(self.device)
            else:
                raise RuntimeError("Gradient argument must be specified for non-scalar tensors.")


        stack = [(self, gradient)]
        visited = set()
    
        while stack:
            tensor, grad = stack.pop()
            
            if tensor.grad is None:
                tensor.grad = grad
            else:
                tensor.grad += grad

            # Propagate gradients to inputs if not a leaf tensor
            if tensor.grad_fn is not None:
                grads = tensor.grad_fn.backward(grad)
                for tensor, grad in zip(tensor.grad_fn.input, grads):
                    if isinstance(tensor, Tensor) and tensor not in visited:
                        stack.append((tensor, grad))
                        visited.add(tensor)

    def zero_grad(self):
        self.grad = None

    def detach(self):
        self.grad = None
        self.grad_fn = None

    def __getitem__(self, indices):
        if isinstance(indices, int):
            indices = [indices]
        if len(indices) != self.ndim:
            raise ValueError("Number of indices must match the number of dimensions")
        
        Tensor._C.get_item.argtypes = [ctypes.POINTER(CTensor), ctypes.POINTER(ctypes.c_int)]
        Tensor._C.get_item.restype = ctypes.c_float
                                           
        indices = (ctypes.c_int * len(indices))(*indices)
        value = Tensor._C.get_item(self.tensor, indices)  

        return value 
    
    def __str__(self):
        def print_recursively(tensor, depth, index):
            if depth == tensor.ndim - 1:
                result = ""
                for i in range(tensor.shape[-1]):
                    index[-1] = i
                    result += str(tensor[tuple(index)]) + ", "
                return result.strip()
            else:
                result = ""
                if depth > 0:
                    result += "\n" + " " * ((depth - 1) * 4)
                for i in range(tensor.shape[depth]):
                    index[depth] = i
                    result += "["
                    result += print_recursively(tensor, depth + 1, index) + "],"
                    if i < tensor.shape[depth] - 1:
                        result += "\n" + " " * (depth * 4)
                return result.strip(",")

        index = [0] * self.ndim
        result = "tensor(["
        result += print_recursively(self, 0, index)
        result += f"""], device="{self.device}", requires_grad={self.requires_grad})"""
        return result

    def __repr__(self):
        return self.__str__()
    
    def __add__(self, other):
        if isinstance(other, (int, float)):
            other = other * self.ones_like()

        broadcasted_shape_add = []

        # Function to determine if broadcasting is needed and get the broadcasted shape
        def broadcast_shape(shape1, shape2):
            if shape1 == shape2:
                return shape1, False
            
            max_len = max(len(shape1), len(shape2))
            shape1 = [1] * (max_len - len(shape1)) + shape1
            shape2 = [1] * (max_len - len(shape2)) + shape2

            
            for dim1, dim2 in zip(shape1, shape2):
                if dim1 != dim2 and dim1 != 1 and dim2 != 1:
                    raise ValueError("Shapes are not compatible for broadcasting")
                broadcasted_shape_add.append(max(dim1, dim2))
            return broadcasted_shape_add, True

        broadcasted_shape_add, needs_broadcasting = broadcast_shape(self.shape, other.shape)

        if needs_broadcasting:
            # Call add_broadcasted_tensor if broadcasting is needed
            if other.ndim == self.ndim - 1:
                other = other.reshape([1] + other.shape)
            
            elif self.ndim == other.ndim - 1:
                self = self.reshape([1] + self.shape)
                
            
                
            Tensor._C.add_broadcasted_tensor.argtypes = [ctypes.POINTER(CTensor), ctypes.POINTER(CTensor)]
            Tensor._C.add_broadcasted_tensor.restype = ctypes.POINTER(CTensor)

            result_tensor_ptr = Tensor._C.add_broadcasted_tensor(self.tensor, other.tensor)

            result_data = Tensor()
            result_data.tensor = result_tensor_ptr
            result_data.shape = broadcasted_shape_add.copy()
            result_data.ndim = len(broadcasted_shape_add)

            result_data.device = self.device
            result_data.numel = 1
            for s in result_data.shape:
                result_data.numel *= s

            result_data.requires_grad = self.requires_grad or other.requires_grad
            if result_data.requires_grad:
                result_data.grad_fn = AddBroadcastedBackward(self, other)
        
        else:
            # Call add_tensor if shapes are identical
            Tensor._C.add_tensor.argtypes = [ctypes.POINTER(CTensor), ctypes.POINTER(CTensor)]
            Tensor._C.add_tensor.restype = ctypes.POINTER(CTensor)

            result_tensor_ptr = Tensor._C.add_tensor(self.tensor, other.tensor)

            result_data = Tensor()
            result_data.tensor = result_tensor_ptr
            result_data.shape = self.shape.copy()
            result_data.ndim = self.ndim

            result_data.device = self.device
            result_data.numel = self.numel  # Update this to calculate the correct number of elements if broadcasting

            result_data.requires_grad = self.requires_grad or other.requires_grad
            if result_data.requires_grad:
                result_data.grad_fn = AddBackward(self, other)

        return result_data

    
    def __radd__(self, other):
        if isinstance(other, (int, float)):
            other = other * self.ones_like()
        
        if self.shape != other.shape:
            raise ValueError("Tensors must have the same shape for addition")
        
        Tensor._C.add_tensor.argtypes = [ctypes.POINTER(CTensor), ctypes.POINTER(CTensor)]
        Tensor._C.add_tensor.restype = ctypes.POINTER(CTensor)

        result_tensor_ptr = Tensor._C.add_tensor(other.tensor, self.tensor)

        result_data = Tensor()
        result_data.tensor = result_tensor_ptr
        result_data.shape = self.shape.copy()
        result_data.ndim = self.ndim
        result_data.device = self.device
        result_data.numel = self.numel

        result_data.requires_grad = self.requires_grad or other.requires_grad
        if result_data.requires_grad:
            result_data.grad_fn = AddBackward(other, self)

        return result_data
    
    def __sub__(self, other):
        if isinstance(other, (int, float)):
            other = other * self.ones_like()

        broadcasted_shape_sub = []

        # Function to determine if broadcasting is needed and get the broadcasted shape
        def broadcast_shape(shape1, shape2):
            if shape1 == shape2:
                return shape1, False
            
            max_len = max(len(shape1), len(shape2))
            shape1 = [1] * (max_len - len(shape1)) + shape1
            shape2 = [1] * (max_len - len(shape2)) + shape2

            
            for dim1, dim2 in zip(shape1, shape2):
                if dim1 != dim2 and dim1 != 1 and dim2 != 1:
                    raise ValueError("Shapes are not compatible for broadcasting")
                broadcasted_shape_sub.append(max(dim1, dim2))
            return broadcasted_shape_sub, True

        broadcasted_shape_sub, needs_broadcasting = broadcast_shape(self.shape, other.shape)

        if needs_broadcasting:
            if other.ndim == self.ndim - 1:
                other = other.reshape([1] + other.shape)
            
            elif self.ndim == other.ndim - 1:
                self = self.reshape([1] + self.shape)

            # Call add_broadcasted_tensor if broadcasting is needed
            Tensor._C.sub_broadcasted_tensor.argtypes = [ctypes.POINTER(CTensor), ctypes.POINTER(CTensor)]
            Tensor._C.sub_broadcasted_tensor.restype = ctypes.POINTER(CTensor)

            result_tensor_ptr = Tensor._C.sub_broadcasted_tensor(self.tensor, other.tensor)

            result_data = Tensor()
            result_data.tensor = result_tensor_ptr
            result_data.shape = broadcasted_shape_sub.copy()
            result_data.ndim = len(broadcasted_shape_sub)

            result_data.device = self.device
            result_data.numel = self.numel  # Update this to calculate the correct number of elements if broadcasting

            result_data.requires_grad = self.requires_grad or other.requires_grad
            if result_data.requires_grad:
                result_data.grad_fn = SubBroadcastedBackward(self, other)
        
        else:
            # Call add_tensor if shapes are identical
            Tensor._C.sub_tensor.argtypes = [ctypes.POINTER(CTensor), ctypes.POINTER(CTensor)]
            Tensor._C.sub_tensor.restype = ctypes.POINTER(CTensor)

            result_tensor_ptr = Tensor._C.sub_tensor(self.tensor, other.tensor)

            result_data = Tensor()
            result_data.tensor = result_tensor_ptr
            result_data.shape = self.shape.copy()
            result_data.ndim = self.ndim

            result_data.device = self.device
            result_data.numel = self.numel  # Update this to calculate the correct number of elements if broadcasting

            result_data.requires_grad = self.requires_grad or other.requires_grad
            if result_data.requires_grad:
                result_data.grad_fn = SubBackward(self, other)

        return result_data
    
    def __rsub__(self, other):
        if isinstance(other, (int, float)):
            other = other * self.ones_like()

        if self.shape != other.shape:
            raise ValueError("Tensors must have the same shape for subtraction")
        
        Tensor._C.sub_tensor.argtypes = [ctypes.POINTER(CTensor), ctypes.POINTER(CTensor)]
        Tensor._C.sub_tensor.restype = ctypes.POINTER(CTensor)

        result_tensor_ptr = Tensor._C.sub_tensor(other.tensor, self.tensor)

        result_data = Tensor()
        result_data.tensor = result_tensor_ptr
        result_data.shape = self.shape.copy()
        result_data.ndim = self.ndim
        result_data.device = self.device
        result_data.numel = self.numel

        result_data.requires_grad = self.requires_grad or other.requires_grad
        if result_data.requires_grad:
            result_data.grad_fn = SubBackward(other, self)

        return result_data
    
    def __mul__(self, other):
        if isinstance(other, (int, float)):
            result_data = Tensor()
            result_data.shape = self.shape.copy()
            result_data.ndim = self.ndim
            result_data.device = self.device
            result_data.numel = self.numel

            Tensor._C.scalar_mul_tensor.argtypes = [ctypes.POINTER(CTensor), ctypes.c_float]
            Tensor._C.scalar_mul_tensor.restype = ctypes.POINTER(CTensor)

            result_data.tensor = Tensor._C.scalar_mul_tensor(self.tensor, ctypes.c_float(other))

            result_data.requires_grad = self.requires_grad
            if result_data.requires_grad:
                result_data.grad_fn = ScalarMulBackward(self, other)

            return result_data
        elif isinstance(other, Tensor):
            if self.shape != other.shape:
                raise ValueError("Tensors must have the same shape for element-wise multiplication")

            Tensor._C.elementwise_mul_tensor.argtypes = [ctypes.POINTER(CTensor), ctypes.POINTER(CTensor)]
            Tensor._C.elementwise_mul_tensor.restype = ctypes.POINTER(CTensor)

            result_tensor_ptr = Tensor._C.elementwise_mul_tensor(self.tensor, other.tensor)

            result_data = Tensor()
            result_data.tensor = result_tensor_ptr
            result_data.shape = self.shape.copy()
            result_data.ndim = self.ndim
            result_data.device = self.device
            result_data.numel = self.numel

            result_data.requires_grad = self.requires_grad or other.requires_grad
            if result_data.requires_grad:
                result_data.grad_fn = ElementwiseMulBackward(self, other)

            return result_data
        else:
            raise TypeError("Unsupported operand type(s) for *: '{}' and '{}'".format(type(self), type(other)))

    def __rmul__(self, other):
        return self.__mul__(other)

    def __neg__(self):
        return self.__mul__(-1)  

    def __pos__(self):
        return self  
    
    def __matmul__(self, other):
        if self.ndim < 3 and other.ndim == 3:
            #broadcasted 2D x 3D matmul

            Tensor._C.broadcasted_batched_matmul_tensor.argtypes = [ctypes.POINTER(CTensor), ctypes.POINTER(CTensor)]
            Tensor._C.broadcasted_batched_matmul_tensor.restype = ctypes.POINTER(CTensor)
            
            result_tensor_ptr = Tensor._C.broadcasted_batched_matmul_tensor(self.tensor, other.tensor)

            result_data = Tensor()
            result_data.tensor = result_tensor_ptr
            result_data.shape = [other.shape[0], self.shape[0], other.shape[2]]
            result_data.ndim = 3
            result_data.device = self.device
            result_data.numel = 1
            for s in result_data.shape:
                result_data.numel *= s


        elif self.ndim == 3 and other.ndim == 3:
            #broadcasted 3D x 3D matmul

            Tensor._C.batched_matmul_tensor.argtypes = [ctypes.POINTER(CTensor), ctypes.POINTER(CTensor)]
            Tensor._C.batched_matmul_tensor.restype = ctypes.POINTER(CTensor)
            
            result_tensor_ptr = Tensor._C.batched_matmul_tensor(self.tensor, other.tensor)

            result_data = Tensor()
            result_data.tensor = result_tensor_ptr
            result_data.shape = [other.shape[0], self.shape[1], other.shape[2]]
            result_data.ndim = 3
            result_data.device = self.device
            result_data.numel = 1
            for s in result_data.shape:
                result_data.numel *= s
        else:
            #2D matmul
            if self.ndim != 2 or other.ndim != 2:
                raise ValueError("Matrix multiplication requires 2D tensors")

            if self.shape[1] != other.shape[0]:
                raise ValueError("Incompatible shapes for matrix multiplication")

            Tensor._C.matmul_tensor.argtypes = [ctypes.POINTER(CTensor), ctypes.POINTER(CTensor)]
            Tensor._C.matmul_tensor.restype = ctypes.POINTER(CTensor)

            result_tensor_ptr = Tensor._C.matmul_tensor(self.tensor, other.tensor)

            result_data = Tensor()
            result_data.tensor = result_tensor_ptr
            result_data.shape = [self.shape[0], other.shape[1]]
            result_data.ndim = 2
            result_data.device = self.device
            result_data.numel = 1
            for s in result_data.shape:
                result_data.numel *= s

        result_data.requires_grad = self.requires_grad or other.requires_grad
        if result_data.requires_grad:
            result_data.grad_fn = MatmulBackward(self, other)

        return result_data

    def __pow__(self, other):
        other = float(other)
        Tensor._C.tensor_pow_scalar.argtypes = [ctypes.POINTER(CTensor), ctypes.c_float]
        Tensor._C.tensor_pow_scalar.restype = ctypes.POINTER(CTensor)

        result_tensor_ptr = Tensor._C.tensor_pow_scalar(self.tensor, ctypes.c_float(other))

        result_data = Tensor()
        result_data.tensor = result_tensor_ptr
        result_data.shape = self.shape.copy()
        result_data.ndim = self.ndim
        result_data.device = self.device
        result_data.numel = self.numel

        result_data.requires_grad = self.requires_grad
        if result_data.requires_grad:
            result_data.grad_fn = PowBackward(self, other)
        
        return result_data
    
    def __rpow__(self, other):
        other = float(other)
        Tensor._C.scalar_pow_tensor.argtypes = [ctypes.c_float, ctypes.POINTER(CTensor)]
        Tensor._C.scalar_pow_tensor.restype = ctypes.POINTER(CTensor)

        result_tensor_ptr = Tensor._C.scalar_pow_tensor(ctypes.c_float(other), self.tensor)

        result_data = Tensor()
        result_data.tensor = result_tensor_ptr
        result_data.shape = self.shape.copy()
        result_data.ndim = self.ndim
        result_data.device = self.device
        result_data.numel = self.numel

        result_data.requires_grad = self.requires_grad
        if result_data.requires_grad:
            result_data.grad_fn = PowBackward(other, self)

        return result_data

    def __truediv__(self, other):
        if isinstance(other, (int, float)):
            other = float(other)
            Tensor._C.tensor_div_scalar.argtypes = [ctypes.POINTER(CTensor), ctypes.c_float]
            Tensor._C.tensor_div_scalar.restype = ctypes.POINTER(CTensor)

            result_tensor_ptr = Tensor._C.tensor_div_scalar(self.tensor, ctypes.c_float(other))

            result_data = Tensor()
            result_data.tensor = result_tensor_ptr
            result_data.shape = self.shape.copy()
            result_data.ndim = self.ndim
            result_data.device = self.device
            result_data.numel = self.numel
            
            result_data.requires_grad = self.requires_grad
            if result_data.requires_grad:
                result_data.grad_fn = DivisionBackward(self, other)
        
        elif isinstance(self, Tensor) and isinstance(other, Tensor):
            if other.numel == 1:
                return self.__truediv__(other.tensor.contents.data[0])
            
            Tensor._C.tensor_div_tensor.argtypes = [ctypes.POINTER(CTensor), ctypes.POINTER(CTensor)]
            Tensor._C.tensor_div_tensor.restype = ctypes.POINTER(CTensor)

            result_tensor_ptr = Tensor._C.tensor_div_tensor(self.tensor, other.tensor)

            result_data = Tensor()
            result_data.tensor = result_tensor_ptr
            result_data.shape = self.shape.copy()
            result_data.ndim = self.ndim
            result_data.device = self.device
            result_data.numel = self.numel

            result_data.requires_grad = self.requires_grad or other.requires_grad
            if result_data.requires_grad:
                result_data.grad_fn = DivisionBackward(self, other)

        return result_data
        
        
    
    def __rtruediv__(self, other):
        other = float(other)

        Tensor._C.scalar_div_tensor.argtypes = [ctypes.c_float, ctypes.POINTER(CTensor)]
        Tensor._C.scalar_div_tensor.restype = ctypes.POINTER(CTensor)

        result_tensor_ptr = Tensor._C.scalar_div_tensor(ctypes.c_float(other), self.tensor)        

        result_data = Tensor()
        result_data.tensor = result_tensor_ptr
        result_data.shape = self.shape.copy()
        result_data.ndim = self.ndim
        result_data.device = self.device
        result_data.numel = self.numel

        result_data.requires_grad = self.requires_grad
        if result_data.requires_grad:
            result_data.grad_fn = DivisionBackward(other, self)

        return result_data

    
    def equal(self, other):

        if isinstance(other, Tensor) and other.numel == 1:
            # other is a single value tensor
            other = self.zeros_like() + other
            return self.equal(other)

        if not isinstance(other, Tensor):
            # other is a single value
            if isinstance(other, (int, float)):
                other = self.zeros_like() + other

                return self.equal(other)
            else:
                return False
        
        broadcasted_shape_add = []

        # Function to determine if broadcasting is needed and get the broadcasted shape
        def broadcast_shape(shape1, shape2):
            if shape1 == shape2:
                return shape1, False
            
            max_len = max(len(shape1), len(shape2))
            shape1 = [1] * (max_len - len(shape1)) + shape1
            shape2 = [1] * (max_len - len(shape2)) + shape2

            
            for dim1, dim2 in zip(shape1, shape2):
                if dim1 != dim2 and dim1 != 1 and dim2 != 1:
                    raise ValueError("Shapes are not compatible for broadcasting")
                broadcasted_shape_add.append(max(dim1, dim2))
            return broadcasted_shape_add, True

        broadcasted_shape_add, needs_broadcasting = broadcast_shape(self.shape, other.shape)
        
        if needs_broadcasting:
            if other.ndim == self.ndim - 1:
                other = other.reshape([1] + other.shape)
            
            elif self.ndim == other.ndim - 1:
                self = self.reshape([1] + self.shape)

            # Call equal_broadcasted_tensor if broadcasting is needed
            Tensor._C.equal_broadcasted_tensor.argtypes = [ctypes.POINTER(CTensor), ctypes.POINTER(CTensor)]
            Tensor._C.equal_broadcasted_tensor.restype = ctypes.POINTER(CTensor)

            result_tensor_ptr = Tensor._C.equal_broadcasted_tensor(self.tensor, other.tensor)

            result_data = Tensor()
            result_data.tensor = result_tensor_ptr
            result_data.shape = broadcasted_shape_add.copy()
            result_data.ndim = len(broadcasted_shape_add)

            result_data.device = self.device
            result_data.numel = 1
            for s in result_data.shape:
                result_data.numel *= s

        else:
            Tensor._C.equal_tensor.argtypes = [ctypes.POINTER(CTensor), ctypes.POINTER(CTensor)]
            Tensor._C.equal_tensor.restype = ctypes.POINTER(CTensor)

            result_tensor_ptr = Tensor._C.equal_tensor(self.tensor, other.tensor)

            result_data = Tensor()
            result_data.tensor = result_tensor_ptr
            result_data.shape = self.shape.copy()
            result_data.ndim = self.ndim
            result_data.device = self.device
            result_data.numel = self.numel
            
        return result_data

    def log(self):
        Tensor._C.log_tensor.argtypes = [ctypes.POINTER(CTensor)]
        Tensor._C.log_tensor.restype = ctypes.POINTER(CTensor)

        result_tensor_ptr = Tensor._C.log_tensor(self.tensor)

        result_data = Tensor()
        result_data.tensor = result_tensor_ptr
        result_data.shape = self.shape.copy()
        result_data.ndim = self.ndim
        result_data.device = self.device
        result_data.numel = self.numel

        result_data.requires_grad = self.requires_grad
        if result_data.requires_grad:
            result_data.grad_fn = LogBackward(self)

        return result_data
    
    def sum(self, axis=None, keepdim=False):
        if axis is not None and axis < 0:
            axis = self.ndim + axis
            
        if axis == None:
            axis = -1

        if axis > self.ndim - 1:
            raise ValueError(f"Error: axis argument {axis} cannot be higher than tensor dimension {self.ndim}")

        Tensor._C.sum_tensor.argtypes = [ctypes.POINTER(CTensor), ctypes.c_int, ctypes.c_bool]
        Tensor._C.sum_tensor.restype = ctypes.POINTER(CTensor)

        result_tensor_ptr = Tensor._C.sum_tensor(self.tensor, axis, keepdim)

        result_data = Tensor()
        result_data.tensor = result_tensor_ptr

        if axis == -1:
            if keepdim:
                result_data.ndim = self.ndim
                result_data.shape = [1] * self.ndim

            else:
                result_data.shape = [1]
                result_data.ndim = 1
        else:
            if keepdim:
                result_data.shape = self.shape[:axis] + [1] + self.shape[axis+1:]
            else:
                result_data.shape = self.shape[:axis] + self.shape[axis+1:]
            result_data.ndim = len(result_data.shape)

        result_data.device = self.device
        result_data.numel = 1
        for s in result_data.shape:
            result_data.numel *= s

        result_data.requires_grad = self.requires_grad
        if result_data.requires_grad:
            result_data.grad_fn = SumBackward(self, axis, keepdim=keepdim)

        return result_data
    
    def max(self, axis=None, keepdim=False):
        if axis is not None and axis < 0:
            axis = self.ndim + axis
            
        if axis == None:
            axis = -1

        if axis > self.ndim - 1:
            raise ValueError(f"Error: axis argument {axis} cannot be higher than tensor dimension {self.ndim}")

        Tensor._C.max_tensor.argtypes = [ctypes.POINTER(CTensor), ctypes.c_int, ctypes.c_bool]
        Tensor._C.max_tensor.restype = ctypes.POINTER(CTensor)

        result_tensor_ptr = Tensor._C.max_tensor(self.tensor, axis, keepdim)

        result_data = Tensor()
        result_data.tensor = result_tensor_ptr

        if axis == -1:
            if keepdim:
                result_data.ndim = self.ndim
                result_data.shape = [1] * self.ndim

            else:
                result_data.shape = [1]
                result_data.ndim = 1
        else:
            if keepdim:
                result_data.shape = self.shape[:axis] + [1] + self.shape[axis+1:]
            else:
                result_data.shape = self.shape[:axis] + self.shape[axis+1:]
            result_data.ndim = len(result_data.shape)

        result_data.device = self.device
        result_data.numel = 1
        for s in result_data.shape:
            result_data.numel *= s

        result_data.requires_grad = self.requires_grad
        if result_data.requires_grad:
            result_data.grad_fn = MaxBackward(self, axis, keepdim=keepdim)

        return result_data
    
    def min(self, axis=None, keepdim=False):
        if axis is not None and axis < 0:
            axis = self.ndim + axis

        if axis == None:
            axis = -1

        if axis > self.ndim - 1:
            raise ValueError(f"Error: axis argument {axis} must be smaller than tensor dimension {self.ndim}")
        
        Tensor._C.min_tensor.argtypes = [ctypes.POINTER(CTensor), ctypes.c_int, ctypes.c_bool]
        Tensor._C.min_tensor.restype = ctypes.POINTER(CTensor)

        result_tensor_ptr = Tensor._C.min_tensor(self.tensor, axis, keepdim)

        result_data = Tensor()
        result_data.tensor = result_tensor_ptr

        if axis == -1:
            if keepdim:
                result_data.ndim = self.ndim
                result_data.shape = [1] * self.ndim

            else:
                result_data.shape = [1]
                result_data.ndim = 1
        else:
            if keepdim:
                result_data.shape = self.shape[:axis] + [1] + self.shape[axis+1:]
            else:
                result_data.shape = self.shape[:axis] + self.shape[axis+1:]
            result_data.ndim = len(result_data.shape)

        result_data.device = self.device
        result_data.numel = 1
        for s in result_data.shape:
            result_data.numel *= s

        result_data.requires_grad = self.requires_grad
        if result_data.requires_grad:
            result_data.grad_fn = MinBackward(self, axis, keepdim=keepdim)

        return result_data

    
    def sin(self):
        Tensor._C.sin_tensor.argtypes = [ctypes.POINTER(CTensor)]
        Tensor._C.sin_tensor.restype = ctypes.POINTER(CTensor)

        result_tensor_ptr = Tensor._C.sin_tensor(self.tensor)

        result_data = Tensor()
        result_data.tensor = result_tensor_ptr
        result_data.shape = self.shape.copy()
        result_data.ndim = self.ndim
        result_data.device = self.device
        result_data.numel = self.numel

        result_data.requires_grad = self.requires_grad
        if result_data.requires_grad:
            result_data.grad_fn = SinBackward(self)
        
        return result_data
    
    def cos(self):
        Tensor._C.cos_tensor.argtypes = [ctypes.POINTER(CTensor)]
        Tensor._C.cos_tensor.restype = ctypes.POINTER(CTensor)

        result_tensor_ptr = Tensor._C.cos_tensor(self.tensor)

        result_data = Tensor()
        result_data.tensor = result_tensor_ptr
        result_data.shape = self.shape.copy()
        result_data.ndim = self.ndim
        result_data.device = self.device
        result_data.numel = self.numel

        result_data.requires_grad = self.requires_grad
        if result_data.requires_grad:
            result_data.grad_fn = CosBackward(self)
        
        return result_data
    
    def sigmoid(self):
        Tensor._C.sigmoid_tensor.argtypes = [ctypes.POINTER(CTensor)]
        Tensor._C.sigmoid_tensor.restype = ctypes.POINTER(CTensor)

        result_tensor_ptr = Tensor._C.sigmoid_tensor(self.tensor)

        result_data = Tensor()
        result_data.tensor = result_tensor_ptr
        result_data.shape = self.shape.copy()
        result_data.ndim = self.ndim
        result_data.device = self.device
        result_data.numel = self.numel

        result_data.requires_grad = self.requires_grad
        if result_data.requires_grad:
            result_data.grad_fn = SigmoidBackward(self)
        
        return result_data
    

    
    def transpose(self, axis1, axis2):
        if axis1 < 0:
            axis1 = self.ndim + axis1
        if axis2 < 0:
            axis2 = self.ndim + axis2

        Tensor._C.transpose_axes_tensor.argtypes = [ctypes.POINTER(CTensor), ctypes.c_int, ctypes.c_int]
        Tensor._C.transpose_axes_tensor.restype = ctypes.POINTER(CTensor)

        result_tensor_ptr = Tensor._C.transpose_axes_tensor(self.tensor, axis1, axis2)

        result_data = Tensor()
        result_data.tensor = result_tensor_ptr
        result_data.shape = self.shape.copy()
        result_data.shape[axis1] = self.shape[axis2]
        result_data.shape[axis2] = self.shape[axis1]
        result_data.ndim = self.ndim
        result_data.device = self.device
        result_data.numel = self.numel

        result_data.requires_grad = self.requires_grad
        if result_data.requires_grad:
            result_data.grad_fn = TransposeBackward(self, axis1, axis2)

        return result_data
    
    @property
    def T(self):
        Tensor._C.transpose_tensor.argtypes = [ctypes.POINTER(CTensor)]
        Tensor._C.transpose_tensor.restype = ctypes.POINTER(CTensor)

        result_tensor_ptr = Tensor._C.transpose_tensor(self.tensor)

        result_data = Tensor()
        result_data.tensor = result_tensor_ptr
        result_data.shape = self.shape.copy()[::-1]
        result_data.ndim = self.ndim
        result_data.device = self.device
        result_data.numel = self.numel

        result_data.requires_grad = self.requires_grad
        if result_data.requires_grad:
            result_data.grad_fn = TBackward(self)

        return result_data
    
    def detach(self):
        self.grad = None
        self.grad_fn = None

        return self        


from .parameter import Parameter
from collections import OrderedDict
from abc import ABC
import pickle
import json
import inspect
import warnings

class Module(ABC):
    """
    Abstract class for modules
    """
    def __init__(self):
        self._modules = OrderedDict()
        self._params = OrderedDict()
        self._grads = OrderedDict()
        self.training = True

    def forward(self, *inputs, **kwargs):
        raise NotImplementedError

    def __call__(self, *inputs, **kwargs):
        return self.forward(*inputs, **kwargs)

    def train(self):
        self.training = True
        for param in self.parameters():
            param.requires_grad = True

    def eval(self):
        self.training = False
        for param in self.parameters():
            param.requires_grad = False

    def parameters(self):
        for name, value in inspect.getmembers(self):
            if isinstance(value, Parameter):
                yield self, name, value
            elif isinstance(value, Module):
                yield from value.parameters()

    def modules(self):
        yield from self._modules.values()

    def gradients(self):
        for module in self.modules():
            yield module._grads

    def zero_grad(self):
        for _, _, parameter in self.parameters():
            parameter.zero_grad()

    def to(self, device):
        for module, name, _ in self.parameters():
            parameter = getattr(module, name)
            parameter = parameter.to(device)
            setattr(module, name, parameter)

        return self
    
    def state_dict(self):
        state = OrderedDict()
        for i, param in enumerate(self.parameters()):
            state[f'param{i}'] = param.tolist()
        return state
    
    def load_state(self, state_dict):
        for i, param in self.parameters():
            data = state_dict[f'param{i}']
            if param.shape != data.shape:
                warnings.warn(f"The 'state_dict' shape does not match model's parameter shape. "
                              f"Got {data.shape}, expected {param.shape}.")
            param.data = Parameter(data=data)

    def save(self, filename='model.pickle'):
        with open(filename, 'wb') as f:
            pickle.dump(self, f)

    def save_dict(self, filename='state_dict.json'):
        state = self.state_dict()
        with open(filename, 'w') as f:
            json.dump(state, f)

    def inner_repr(self):
        return ""

    def __repr__(self):
        string = f"{self.get_name()}("
        tab = "   "
        modules = self._modules
        if modules == {}:
            string += f'\n{tab}(parameters): {self.inner_repr()}'
        else:
            for key, module in modules.items():
                string += f"\n{tab}({key}): {module.get_name()}({module.inner_repr()})"
        return f'{string}\n)'
    
    def get_name(self):
        return self.__class__.__name__
    
    def __setattr__(self, key, value):
        self.__dict__[key] = value

        if isinstance(value, Module):
            self._modules[key] = value
        elif isinstance(value, Parameter):
            self._params[key] = value


import unittest
import norch
from norch.utils import utils_unittests as utils
import torch
import sys
import os

class TestTensorOperations(unittest.TestCase):

    def setUp(self):
        self.device = os.environ.get('device')
        if self.device is None or self.device != 'cuda':
            self.device = 'cpu'

        print(f"Running tests on: {self.device}")

    def test_creation_and_conversion(self):
        """
        Test creation and convertion of norch tensor to pytorch
        """
        norch_tensor = norch.Tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        torch_tensor = utils.to_torch(norch_tensor)
        self.assertTrue(torch.is_tensor(torch_tensor))

    def test_addition(self):
        """
        Test addition two tensors: tensor1 + tensor2
        """
        norch_tensor1 = norch.Tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        norch_tensor2 = norch.Tensor([[[1, 1], [1, 1]], [[1, 1], [1, 1]]]).to(self.device)
        norch_result = norch_tensor1 + norch_tensor2
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_tensor1 = torch.tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        torch_tensor2 = torch.tensor([[[1, 1], [1, 1]], [[1, 1], [1, 1]]]).to(self.device)
        torch_expected = torch_tensor1 + torch_tensor2

        self.assertTrue(utils.compare_torch(torch_result, torch_expected))

    def test_addition_broadcasted(self):
        """
        Test addition of two tensors with broadcasting: tensor1 + tensor2
        """
        norch_tensor1 = norch.Tensor([[[1, 2, 3], [4, 5, 6]]]).to(self.device)  # Shape (1, 2, 3)
        norch_tensor2 = norch.Tensor([1, 1, 1]).to(self.device)  # Shape (3)
        norch_result = norch_tensor1 + norch_tensor2
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_tensor1 = torch.tensor([[[1, 2, 3], [4, 5, 6]]]).to(self.device)  # Shape (1, 2, 3)
        torch_tensor2 = torch.tensor([1, 1, 1]).to(self.device)  # Shape (3)
        torch_expected = torch_tensor1 + torch_tensor2

        self.assertTrue(utils.compare_torch(torch_result, torch_expected))

        norch_tensor1 = norch.Tensor([[[10, 10], [-4, -4]], [[5., 6], [7, 8]]]).to(self.device)  # Shape (1, 2, 3)
        norch_tensor2 = norch.Tensor([[10, 10], [5, 6]]).to(self.device)  # Shape (3)
        norch_result = norch_tensor1 + norch_tensor2
        torch_result = utils.to_torch(norch_result).to(self.device)
        
        torch_tensor1 = torch.tensor([[[10, 10], [-4, -4]], [[5., 6], [7, 8]]]).to(self.device)  # Shape (1, 2, 3)
        torch_tensor2 = torch.tensor([[[10, 10], [5, 6]]]).to(self.device)  # Shape (3)
        torch_expected = torch_tensor1 + torch_tensor2

        self.assertTrue(utils.compare_torch(torch_result, torch_expected))

        # reversed order broadcasting
        norch_tensor1 = norch.Tensor([[0, 2]]).to(self.device) 
        norch_tensor2 = norch.Tensor([[3, 4], [5, -1]]).to(self.device) 
        norch_result = norch_tensor1 + norch_tensor2
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_tensor1 = torch.tensor([[0, 2]]).to(self.device)  
        torch_tensor2 = torch.tensor([[3, 4], [5, -1]]).to(self.device) 
        torch_expected = torch_tensor1 + torch_tensor2

        self.assertTrue(utils.compare_torch(torch_result, torch_expected))

        norch_result = norch_tensor2 + norch_tensor1
        torch_expected = torch_tensor2 + torch_tensor1

        self.assertTrue(utils.compare_torch(torch_result, torch_expected))



    def test_subtraction(self):
        """
        Test subtraction of two tensors: tensor1 - tensor2
        """
        norch_tensor1 = norch.Tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        norch_tensor2 = norch.Tensor([[[1, 1], [1, 1]], [[1, 1], [1, 1]]]).to(self.device)
        norch_result = norch_tensor1 - norch_tensor2
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_tensor1 = torch.tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        torch_tensor2 = torch.tensor([[[1, 1], [1, 1]], [[1, 1], [1, 1]]]).to(self.device)
        torch_expected = torch_tensor1 - torch_tensor2

        self.assertTrue(utils.compare_torch(torch_result, torch_expected))

    def test_broadcasting_subtraction(self):
        """
        Test subtraction of two tensors with broadcasting: tensor1 - tensor2
        """
        norch_tensor1 = norch.Tensor([[[1, 2, 3], [4, 5, 6]]]).to(self.device)  # Shape (1, 2, 3)
        norch_tensor2 = norch.Tensor([1, 1, 1]).to(self.device)  # Shape (3)
        norch_result = norch_tensor1 - norch_tensor2
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_tensor1 = torch.tensor([[[1, 2, 3], [4, 5, 6]]]).to(self.device)  # Shape (1, 2, 3)
        torch_tensor2 = torch.tensor([1, 1, 1]).to(self.device)  # Shape (3)
        torch_expected = torch_tensor1 - torch_tensor2 

        self.assertTrue(utils.compare_torch(torch_result, torch_expected))

        # reversed order broadcasting
        norch_result = norch_tensor2 - norch_tensor1
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_expected = torch_tensor2 - torch_tensor1

        self.assertTrue(utils.compare_torch(torch_result, torch_expected))


    def test_division_by_scalar(self):
        """
        Test division of a tensor by a scalar: tensor / scalar
        """
        norch_tensor = norch.Tensor([[[2, 4], [6, -8]], [[10, 12], [14, 16]]]).to(self.device)
        scalar = 2
        norch_result = norch_tensor / scalar
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_tensor = torch.tensor([[[2, 4], [6, -8]], [[10, 12], [14, 16]]]).to(self.device)
        torch_expected = torch_tensor / scalar

        self.assertTrue(utils.compare_torch(torch_result, torch_expected))

    def test_scalar_division_by_tensor(self):
        """
        Test scalar division by a tensor: scalar / tensor
        """
        scalar = 10
        norch_tensor = norch.Tensor([[[2, 4], [6, -8]], [[10, 12], [14, 16]]]).to(self.device)
        norch_result = scalar / norch_tensor
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_tensor = torch.tensor([[[2, 4], [6, -8]], [[10, 12], [14, 16]]]).to(self.device)
        torch_expected = scalar / torch_tensor

        self.assertTrue(utils.compare_torch(torch_result, torch_expected))

    def test_matrix_multiplication(self):
        """
        Test matrix multiplication: tensor1 @ tensor2
        """
        norch_tensor1 = norch.Tensor([[[1., 2], [3, 4]], [[5, 6], [7, 8]]]).to(self.device)
        norch_tensor2 = norch.Tensor([[[1., 0], [0, 1]], [[-1, 0], [0, -1]]]).to(self.device)
        norch_result = norch_tensor1 @ norch_tensor2
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_tensor1 = torch.tensor([[[1., 2], [3, 4]], [[5, 6], [7, 8]]]).to(self.device)
        torch_tensor2 = torch.tensor([[[1., 0], [0, 1]], [[-1, 0], [0, -1]]]).to(self.device)
        torch_expected = torch_tensor1 @ torch_tensor2

        self.assertTrue(utils.compare_torch(torch_result, torch_expected))

    def test_elementwise_multiplication_by_scalar(self):
        """
        Test elementwise multiplication of a tensor by a scalar: tensor * scalar
        """
        norch_tensor = norch.Tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        scalar = 2
        norch_result = norch_tensor * scalar
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_tensor = torch.tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        torch_expected = torch_tensor * scalar

        self.assertTrue(utils.compare_torch(torch_result, torch_expected))

    def test_elementwise_multiplication_by_tensor(self):
        """
        Test elementwise multiplication of two tensors: tensor1 * tensor2
        """
        norch_tensor1 = norch.Tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        norch_tensor2 = norch.Tensor([[[2, 2], [2, 2]], [[2, 2], [2, 2]]]).to(self.device)
        norch_result = norch_tensor1 * norch_tensor2
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_tensor1 = torch.tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        torch_tensor2 = torch.tensor([[[2, 2], [2, 2]], [[2, 2], [2, 2]]]).to(self.device)
        torch_expected = torch_tensor1 * torch_tensor2

        self.assertTrue(utils.compare_torch(torch_result, torch_expected))

    def test_reshape(self):
        """
        Test reshaping of a tensor: tensor.reshape(shape)
        """
        norch_tensor = norch.Tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        new_shape = [2, 4]
        norch_result = norch_tensor.reshape(new_shape)
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_tensor = torch.tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        torch_expected = torch_tensor.reshape(new_shape)

        self.assertTrue(utils.compare_torch(torch_result, torch_expected))

    def test_unsqueeze(self):
        """
        Test unsqueeze operation on a tensor
        """
        norch_tensor = norch.Tensor([[1, 2], [3, 4]]).to(self.device)
        
        # Unsqueeze at dim=0
        norch_unsqueeze_0 = norch_tensor.unsqueeze(0)
        torch_unsqueeze_0 = utils.to_torch(norch_unsqueeze_0).to(self.device)
        torch_tensor = torch.tensor([[1, 2], [3, 4]]).to(self.device)
        torch_expected_0 = torch_tensor.unsqueeze(0)
        self.assertTrue(utils.compare_torch(torch_unsqueeze_0, torch_expected_0))

        # Unsqueeze at dim=1
        norch_unsqueeze_1 = norch_tensor.unsqueeze(1)
        torch_unsqueeze_1 = utils.to_torch(norch_unsqueeze_1).to(self.device)
        torch_expected_1 = torch_tensor.unsqueeze(1)
        self.assertTrue(utils.compare_torch(torch_unsqueeze_1, torch_expected_1))

        # Unsqueeze at dim=2
        norch_unsqueeze_2 = norch_tensor.unsqueeze(2)
        torch_unsqueeze_2 = utils.to_torch(norch_unsqueeze_2).to(self.device)
        torch_expected_2 = torch_tensor.unsqueeze(2)
        self.assertTrue(utils.compare_torch(torch_unsqueeze_2, torch_expected_2))

        # Unsqueeze at dim=-1
        norch_unsqueeze_neg_1 = norch_tensor.unsqueeze(-1)
        torch_unsqueeze_neg_1 = utils.to_torch(norch_unsqueeze_neg_1).to(self.device)
        torch_expected_neg_1 = torch_tensor.unsqueeze(-1)
        self.assertTrue(utils.compare_torch(torch_unsqueeze_neg_1, torch_expected_neg_1))

        # Unsqueeze at dim=-2
        norch_unsqueeze_neg_2 = norch_tensor.unsqueeze(-2)
        torch_unsqueeze_neg_2 = utils.to_torch(norch_unsqueeze_neg_2).to(self.device)
        torch_expected_neg_2 = torch_tensor.unsqueeze(-2)
        self.assertTrue(utils.compare_torch(torch_unsqueeze_neg_2, torch_expected_neg_2))

    def test_squeeze(self):
        """
        Test squeeze operation on a tensor
        """
        # Create a tensor with some dimensions of size 1
        norch_tensor = norch.Tensor([[[1, 2], [3, 4]]]).to(self.device)  # shape [1, 2, 2]
        
        # Squeeze at dim=0
        norch_squeeze_0 = norch_tensor.squeeze(0)
        torch_squeeze_0 = utils.to_torch(norch_squeeze_0).to(self.device)
        torch_tensor = torch.tensor([[[1, 2], [3, 4]]]).to(self.device)
        torch_expected_0 = torch_tensor.squeeze(0)
        self.assertTrue(utils.compare_torch(torch_squeeze_0, torch_expected_0))

        # Create a tensor with a dimension of size 1 in the middle
        norch_tensor_middle_1 = norch.Tensor([[[1, 2]], [[3, 4]]]).to(self.device)  # shape [2, 1, 2]
        
        # Squeeze at dim=1
        norch_squeeze_1 = norch_tensor_middle_1.squeeze(1)
        torch_squeeze_1 = utils.to_torch(norch_squeeze_1).to(self.device)
        torch_tensor_middle_1 = torch.tensor([[[1, 2]], [[3, 4]]]).to(self.device)
        torch_expected_1 = torch_tensor_middle_1.squeeze(1)
        self.assertTrue(utils.compare_torch(torch_squeeze_1, torch_expected_1))

        # Squeeze at dim=-2 (same as dim=1 in this case)
        norch_squeeze_neg_2 = norch_tensor_middle_1.squeeze(-2)
        torch_squeeze_neg_2 = utils.to_torch(norch_squeeze_neg_2).to(self.device)
        torch_expected_neg_2 = torch_tensor_middle_1.squeeze(-2)
        self.assertTrue(utils.compare_torch(torch_squeeze_neg_2, torch_expected_neg_2))

        # Squeeze all dimensions of size 1 (None)
        norch_tensor_all_1 = norch.Tensor([[[[1, 2], [3, 4]]]]).to(self.device)  # shape [1, 1, 2, 2]
        norch_squeeze_all = norch_tensor_all_1.squeeze()
        torch_squeeze_all = utils.to_torch(norch_squeeze_all).to(self.device)
        torch_tensor_all_1 = torch.tensor([[[[1, 2], [3, 4]]]]).to(self.device)
        torch_expected_all = torch_tensor_all_1.squeeze()
        self.assertTrue(utils.compare_torch(torch_squeeze_all, torch_expected_all))

        # Squeeze no dimensions (no dimensions of size 1)
        norch_tensor_no_1 = norch.Tensor([[1, 2], [3, 4]]).to(self.device)  # shape [2, 2]
        norch_squeeze_none = norch_tensor_no_1.squeeze()
        torch_squeeze_none = utils.to_torch(norch_squeeze_none).to(self.device)
        torch_tensor_no_1 = torch.tensor([[1, 2], [3, 4]]).to(self.device)
        torch_expected_none = torch_tensor_no_1.squeeze()
        self.assertTrue(utils.compare_torch(torch_squeeze_none, torch_expected_none))


    def test_transpose(self):
        """
        Test transposition of a tensor: tensor.transpose(dim1, dim2)
        """
        norch_tensor = norch.Tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        dim1, dim2 = 0, 2
        norch_result = norch_tensor.transpose(dim1, dim2)
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_tensor = torch.tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        torch_expected = torch_tensor.transpose(dim1, dim2)

        self.assertTrue(utils.compare_torch(torch_result, torch_expected))

    def test_logarithm(self):
        """
        Test elementwise logarithm of a tensor: tensor.log()
        """
        norch_tensor = norch.Tensor([[[1, 2], [3, 4]], [[5, 6], [7, 8]]]).to(self.device)
        norch_result = norch_tensor.log()
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_tensor = torch.tensor([[[1, 2], [3, 4]], [[5, 6], [7, 8]]]).to(self.device)
        torch_expected = torch.log(torch_tensor)

        self.assertTrue(utils.compare_torch(torch_result, torch_expected))

    def test_sum(self):
        """
        Test summation of a tensor: tensor.sum()
        """
        norch_tensor = norch.Tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        norch_result = norch_tensor.sum()
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_tensor = torch.tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        torch_expected = torch.sum(torch_tensor)

        self.assertTrue(utils.compare_torch(torch_result, torch_expected))

    def test_sum_axis(self):
        """
        Test summation of a tensor along a specific axis without keeping the dimensions
        """
        norch_tensor = norch.Tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        norch_result = norch_tensor.sum(axis=1)
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_tensor = torch.tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        torch_expected = torch.sum(torch_tensor, dim=1)

        self.assertTrue(utils.compare_torch(torch_result, torch_expected))

        # negative axis

        norch_tensor = norch.Tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        norch_result = norch_tensor.sum(axis=-2)
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_tensor = torch.tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        torch_expected = torch.sum(torch_tensor, dim=-2)

        self.assertTrue(utils.compare_torch(torch_result, torch_expected))


    def test_sum_axis_keepdim(self):
        """
        Test summation of a tensor along a specific axis with keepdim=True
        """
        norch_tensor = norch.Tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        norch_result = norch_tensor.sum(axis=1, keepdim=True)
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_tensor = torch.tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        torch_expected = torch.sum(torch_tensor, dim=1, keepdim=True)
        
        self.assertTrue(utils.compare_torch(torch_result, torch_expected))

    def test_max(self):
        """
        Test max of a tensor: tensor.max()
        """
        norch_tensor = norch.Tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        norch_result = norch_tensor.max()
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_tensor = torch.tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        torch_expected = torch.max(torch_tensor)

        self.assertTrue(utils.compare_torch(torch_result, torch_expected))

    def test_max_axis(self):
        """
        Test max of a tensor along a specific axis without keeping the dimensions
        """
        norch_tensor = norch.Tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        norch_result = norch_tensor.max(axis=1)
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_tensor = torch.tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        torch_expected, _ = torch.max(torch_tensor, dim=1)

        self.assertTrue(utils.compare_torch(torch_result, torch_expected))

        # negative axis

        norch_tensor = norch.Tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        norch_result = norch_tensor.max(axis=-1)
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_tensor = torch.tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        torch_expected, _ = torch.max(torch_tensor, dim=-1)

        self.assertTrue(utils.compare_torch(torch_result, torch_expected))


    def test_max_axis_keepdim(self):
        """
        Test max of a tensor along a specific axis with keepdim=True
        """
        norch_tensor = norch.Tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        norch_result = norch_tensor.max(axis=1, keepdim=True)
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_tensor = torch.tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        torch_expected, _ = torch.max(torch_tensor, dim=1, keepdim=True)
        
        self.assertTrue(utils.compare_torch(torch_result, torch_expected))

    def test_min(self):
        """
        Test min of a tensor: tensor.min()
        """
        norch_tensor = norch.Tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        norch_result = norch_tensor.min()
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_tensor = torch.tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        torch_expected = torch.min(torch_tensor)

        self.assertTrue(utils.compare_torch(torch_result, torch_expected))

    def test_min_axis(self):
        """
        Test min of a tensor along a specific axis without keeping the dimensions
        """
        norch_tensor = norch.Tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        norch_result = norch_tensor.min(axis=1)
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_tensor = torch.tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        torch_expected, _ = torch.min(torch_tensor, dim=1)

        self.assertTrue(utils.compare_torch(torch_result, torch_expected))

        # negative axis

        norch_tensor = norch.Tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        norch_result = norch_tensor.min(axis=-1)
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_tensor = torch.tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        torch_expected, _ = torch.min(torch_tensor, dim=-1)

        self.assertTrue(utils.compare_torch(torch_result, torch_expected))


    def test_min_axis_keepdim(self):
        """
        Test min of a tensor along a specific axis with keepdim=True
        """
        norch_tensor = norch.Tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        norch_result = norch_tensor.min(axis=1, keepdim=True)
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_tensor = torch.tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        torch_expected, _ = torch.min(torch_tensor, dim=1, keepdim=True)
        
        self.assertTrue(utils.compare_torch(torch_result, torch_expected))

    def test_1D_T(self):
        """
        Test transposition of a 1D tensor.
        """
        norch_tensor = norch.Tensor([1, 2, 3, 4]).to(self.device)
        norch_result = norch_tensor.T
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_tensor = torch.tensor([1, 2, 3, 4]).to(self.device)
        torch_expected = torch_tensor.T

        self.assertTrue(utils.compare_torch(torch_result, torch_expected))

    def test_2D_T(self):
        """
        Test transposition of a 2D tensor.
        """
        norch_tensor = norch.Tensor([[1, 2, 3], [4, 5, 6]]).to(self.device)
        norch_result = norch_tensor.T
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_tensor = torch.tensor([[1, 2, 3], [4, 5, 6]]).to(self.device)
        torch_expected = torch_tensor.T

        self.assertTrue(utils.compare_torch(torch_result, torch_expected))

    def test_3D_T(self):
        """
        Test transposition of a tensor: tensor.T
        """
        norch_tensor = norch.Tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        norch_result = norch_tensor.T
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_tensor = torch.tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        torch_expected = torch_tensor.T

        self.assertTrue(utils.compare_torch(torch_result, torch_expected))

    def test_matmul(self):
        """
        Test matrix multiplication: MxP = NxM @ MxP
        """
        # Creating batched tensors for Norch
        norch_tensor1 = norch.Tensor([[1., 2], [3, -4], [5, 6], [7, 8]]).to(self.device)
        norch_tensor2 = norch.Tensor([[2., 3, 1, 0, 4], [5, -1, 2, 3, 0]]).to(self.device)

        norch_result = norch_tensor1 @ norch_tensor2
        torch_result = utils.to_torch(norch_result).to(self.device)

        # Converting to PyTorch tensors for comparison
        torch_tensor1 = torch.tensor([[1., 2], [3, -4], [5, 6], [7, 8]]).to(self.device)
        torch_tensor2 = torch.tensor([[2., 3, 1, 0, 4], [5, -1, 2, 3, 0]]).to(self.device)

        torch_expected = torch.matmul(torch_tensor1, torch_tensor2)

        # Comparing results
        self.assertTrue(utils.compare_torch(torch_result, torch_expected))

    def test_reshape_then_matmul(self):
        """
        Test reshaping a tensor followed by matrix multiplication: (tensor.reshape(shape) @ other_tensor)
        """
        norch_tensor = norch.Tensor([[1., 2], [3, -4], [5, 6], [7, 8]]).to(self.device)
        new_shape = [2, 4]
        norch_reshaped = norch_tensor.reshape(new_shape)
        
        norch_result = norch_reshaped @ norch_tensor
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_tensor = torch.tensor([[1., 2], [3, -4], [5, 6], [7, 8]]).to(self.device)
        torch_expected = torch_tensor.reshape(new_shape) @ torch_tensor

        self.assertTrue(utils.compare_torch(torch_result, torch_expected))

    def test_batched_matmul(self):
        """
        Test batched matrix multiplication: BxMxP = BxNxM @ BxMxP
        """
        B = 3  # Batch size

        # Creating batched tensors for Norch
        norch_tensor1 = norch.Tensor([[[1., 2], [3, -4], [5, 6], [7, 8]] for _ in range(B)]).to(self.device)
        norch_tensor2 = norch.Tensor([[[2., 3, 1, 0, 4], [5, -1, 2, 3, 0]] for _ in range(B)]).to(self.device)

        norch_result = norch_tensor1 @ norch_tensor2
        torch_result = utils.to_torch(norch_result).to(self.device)

        # Converting to PyTorch tensors for comparison
        torch_tensor1 = torch.tensor([[[1., 2], [3, -4], [5, 6], [7, 8]] for _ in range(B)]).to(self.device)
        torch_tensor2 = torch.tensor([[[2., 3, 1, 0, 4], [5, -1, 2, 3, 0]] for _ in range(B)]).to(self.device)

        torch_expected = torch.matmul(torch_tensor1, torch_tensor2)

        # Comparing results
        self.assertTrue(utils.compare_torch(torch_result, torch_expected))


    def test_broadcasted_batched_matmul(self):
        """
        Test broadcasted batched matrix multiplication: BxMxP = NxM @ BxMxP
        """
        B = 3  # Batch size

        # Creating batched tensors for Norch
        norch_tensor1 = norch.Tensor([[1., 2], [3, -4], [5, 6], [7, 8]]).to(self.device)
        norch_tensor2 = norch.Tensor([[[2., 3, 1, 0, 4], [5, -1, 2, 3, 0]] for _ in range(B)]).to(self.device)

        norch_result = norch_tensor1 @ norch_tensor2
        torch_result = utils.to_torch(norch_result).to(self.device)

        # Converting to PyTorch tensors for comparison
        torch_tensor1 = torch.tensor([[1., 2], [3, -4], [5, 6], [7, 8]]).to(self.device)
        torch_tensor2 = torch.tensor([[[2., 3, 1, 0, 4], [5, -1, 2, 3, 0]] for _ in range(B)]).to(self.device)

        torch_expected = torch.matmul(torch_tensor1, torch_tensor2)

        # Comparing results
        self.assertTrue(utils.compare_torch(torch_result, torch_expected))



    def test_transpose_then_matmul(self):
        """
        Test transposing a tensor followed by matrix multiplication: (tensor.transpose(dim1, dim2) @ other_tensor)
        """
        norch_tensor = norch.Tensor([[[1., 2], [3, 4]], [[5, 6], [7, 8]]]).to(self.device)
        dim1, dim2 = 0, 2
        norch_result = norch_tensor.transpose(dim1, dim2) @ norch_tensor
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_tensor = torch.tensor([[[1., 2], [3, 4]], [[5, 6], [7, 8]]]).to(self.device)
        torch_expected = torch_tensor.transpose(dim1, dim2) @ torch_tensor

        self.assertTrue(utils.compare_torch(torch_result, torch_expected))

    def test_add_div_matmul_then_reshape(self):
        """
        Test a combination of operations: (tensor.sum() + other_tensor) / scalar @ another_tensor followed by reshape
        """
        norch_tensor1 = norch.Tensor([[[1., 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        norch_tensor2 = norch.Tensor([[[1., 1], [1, 1]], [[1, 1], [1, 1]]]).to(self.device)
        scalar = 2
        new_shape = [2, 4]
        norch_result = ((norch_tensor1 + norch_tensor2) / scalar) @ norch_tensor1
        norch_result = norch_result.reshape(new_shape)
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_tensor1 = torch.tensor([[[1., 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        torch_tensor2 = torch.tensor([[[1., 1], [1, 1]], [[1, 1], [1, 1]]]).to(self.device)
        torch_expected = ((torch_tensor1 + torch_tensor2) / scalar) @ torch_tensor1
        torch_expected = torch_expected.reshape(new_shape)

        self.assertTrue(utils.compare_torch(torch_result, torch_expected))

    def test_scalar_power_tensor(self):
        """
        Test scalar power of a tensor: scalar ** tensor
        """
        scalar = 3
        norch_tensor = norch.Tensor([[[1, 2.1], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        norch_result = scalar ** norch_tensor
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_tensor = torch.tensor([[[1, 2.1], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        torch_expected = scalar ** torch_tensor

        self.assertTrue(utils.compare_torch(torch_result, torch_expected))

    def test_tensor_power_scalar(self):
        """
        Test tensor power of a scalar: tensor ** scalar
        """
        scalar = 3
        norch_tensor = norch.Tensor([[[1, 2.1], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        norch_result = norch_tensor ** scalar
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_tensor = torch.tensor([[[1, 2.1], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        torch_expected = torch_tensor ** scalar

        self.assertTrue(utils.compare_torch(torch_result, torch_expected))

    def test_tensor_sin(self):
        """
        Test sine function on tensor
        """
        norch_tensor = norch.Tensor([[[0, 30], [45, 60]], [[90, 120], [135, 180]]]).to(self.device)
        norch_result = norch_tensor.sin()
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_tensor = torch.tensor([[[0, 30], [45, 60]], [[90, 120], [135, 180]]]).to(self.device)
        torch_expected = torch.sin(torch_tensor)

        self.assertTrue(utils.compare_torch(torch_result, torch_expected))

    def test_tensor_cos(self):
        """
        Test cosine function on tensor
        """
        norch_tensor = norch.Tensor([[[0, 30], [45, 60]], [[90, 120], [135, 180]]]).to(self.device)
        norch_result = norch_tensor.cos()
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_tensor = torch.tensor([[[0, 30], [45, 60]], [[90, 120], [135, 180]]]).to(self.device)
        torch_expected = torch.cos(torch_tensor)

        self.assertTrue(utils.compare_torch(torch_result, torch_expected))

    def test_equal(self):
        """
        Test equal two tensors: tensor1.equal(tensor2)
        """
        norch_tensor1 = norch.Tensor([[[1, 2], [3, -4]], [[5, 1], [7, 8]]]).to(self.device)
        norch_tensor2 = norch.Tensor([[[1, 1], [1, 1]], [[1, 1], [1, 1]]]).to(self.device)
        norch_result = norch_tensor1.equal(norch_tensor2)
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_tensor1 = torch.tensor([[[1, 2], [3, -4]], [[5, 1], [7, 8]]]).to(self.device)
        torch_tensor2 = torch.tensor([[[1, 1], [1, 1]], [[1, 1], [1, 1]]]).to(self.device)
        torch_expected = (torch_tensor1 == torch_tensor2).float()

        self.assertTrue(utils.compare_torch(torch_result, torch_expected))

    def test_broadcasted_equal(self):
        """
        Test broadcasted equal two tensors: tensor1.equal(tensor2)
        """
        norch_tensor1 = norch.Tensor([[[10, 10], [-4, -4]], [[5., 6], [7, 8]]]).to(self.device)
        norch_tensor2 = norch.Tensor([[[10, 10]], [[5, 6]]]).to(self.device)
        norch_result = norch_tensor1.equal(norch_tensor2)
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_tensor1 = torch.tensor([[[10, 10], [-4, -4]], [[5., 6], [7, 8]]]).to(self.device)
        torch_tensor2 = torch.tensor([[[10, 10]], [[5, 6]]]).to(self.device)
        torch_expected = (torch_tensor1 == torch_tensor2).float()

        self.assertTrue(utils.compare_torch(torch_result, torch_expected))

        norch_tensor1 = norch.Tensor([[[10, 10], [-4, -4]], [[5., 6], [7, 8]]]).to(self.device)
        norch_tensor2 = norch.Tensor([[[10.0,], [-4.0,]],[[6.0,], [8.0,]]]).to(self.device)
        norch_result = norch_tensor1.equal(norch_tensor2)
        torch_result = utils.to_torch(norch_result).to(self.device)

        torch_tensor1 = torch.tensor([[[10, 10], [-4, -4]], [[5., 6], [7, 8]]]).to(self.device)
        torch_tensor2 = torch.tensor([[[10.0,], [-4.0,]],[[6.0,], [8.0,]]]).to(self.device)
        torch_expected = (torch_tensor1 == torch_tensor2).float()

        self.assertTrue(utils.compare_torch(torch_result, torch_expected))
    

    def test_zeros_like(self):
        """
        Test creating a tensor of zeros with the same shape as another tensor.
        """
        norch_tensor = norch.Tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        norch_zeros = norch_tensor.zeros_like()
        torch_zeros_result = utils.to_torch(norch_zeros).to(self.device)

        torch_tensor_expected = torch.tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        torch_zeros_expected = torch.zeros_like(torch_tensor_expected)

        self.assertTrue(utils.compare_torch(torch_zeros_result, torch_zeros_expected))

    def test_ones_like(self):
        """
        Test creating a tensor of ones with the same shape as another tensor.
        """
        norch_tensor = norch.Tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        norch_ones = norch_tensor.ones_like()
        torch_ones_result = utils.to_torch(norch_ones).to(self.device)

        torch_tensor_expected = torch.tensor([[[1, 2], [3, -4]], [[5, 6], [7, 8]]]).to(self.device)
        torch_ones_expected = torch.ones_like(torch_tensor_expected)

        self.assertTrue(utils.compare_torch(torch_ones_result, torch_ones_expected))


if __name__ == '__main__':
    unittest.main()



from ..optimizer import Optimizer
from norch.tensor import Tensor
import time

class SGD(Optimizer):
    def __init__(self, parameters, lr=1e-1, momentum=0):
        super().__init__(parameters)
        self.lr = lr
        self.momentum = momentum
        self._cache = {'velocity': [p.zeros_like() for (_, _, p) in self.parameters]}

    def step(self):
        for i, (module, name, _) in enumerate(self.parameters):
            parameter = getattr(module, name)

            velocity = self._cache['velocity'][i]

            velocity = self.momentum * velocity - self.lr * parameter.grad

            updated_parameter = parameter + velocity

            setattr(module, name, updated_parameter)

            self._cache['velocity'][i] = velocity

            parameter.detach()
            velocity.detach()



import math
import norch
import numpy as np
from norch.autograd.functions import *

def sigmoid(x):
    z = x.sigmoid()
    return z

def softmax(x, dim=None):
    if dim is not None and dim < 0:
        dim = x.ndim + dim
            
    x_max = x.max(axis=dim, keepdim=True)
    exp_x = math.e ** (x - x_max)

    if dim is not None:
        sum_exp_x = exp_x.sum(axis=dim, keepdim=True) + exp_x.zeros_like()
        return exp_x / sum_exp_x
    else:
        sum_exp_x = exp_x.sum()
        return exp_x / sum_exp_x
    
def one_hot_encode(x, num_classes):
    one_hot = [[0] * num_classes for _ in range(x.numel)]

    # Set the appropriate elements to 1
    for i in range(x.numel):
        target_idx = int(x[i])
        one_hot[i][target_idx] = 1

    return norch.Tensor(one_hot)


from .module import Module
from . import functional as F
import math

class Activation(Module):
    """
    Abstract classes for activations
    """
    def __init__(self):
        super(Activation, self).__init__()

    def forward(self, x):
        raise NotImplementedError
    

class Sigmoid(Activation):
    def __init__(self):
        super().__init__()

    def forward(self, x):
        return F.sigmoid(x)
    
class Softmax(Activation):
    def __init__(self, dim):
        super(Softmax, self).__init__()

        self.dim = dim

    def forward(self, x):
        return F.softmax(x, self.dim)


import random
import requests
import tarfile
import zipfile
import shutil
import os

def generate_random_list(shape):
    """
    Generate a list with random numbers and shape 'shape'
    """
    if len(shape) == 0:
        return []
    else:
        inner_shape = shape[1:]
        if len(inner_shape) == 0:
            return [random.uniform(-1, 1) for _ in range(shape[0])]
        else:
            return [generate_random_list(inner_shape) for _ in range(shape[0])]


def download_from_url(url, save_path, chunk_size=128):
    """Download a file from an URL.

    Original answer from https://stackoverflow.com/questions/9419162/download-returned-zip-file-from-url

    Args:
        url (str): path to the URL.
        save_path (str): path to the saving directory.
        chunk_size (int): download chunk.

    Returns:
        None
    """
    response = requests.get(url, stream=True)
    total = response.headers.get('content-length')
    with open(save_path, 'wb') as f:
        if total is None:
            f.write(response.content)
        else:
            downloaded = 0
            total = int(total)
            for data in response.iter_content(chunk_size=max(int(total / 1000), 1024 * 1024)):
                downloaded += len(data)
                f.write(data)
                progress_bar(downloaded, total, "Downloading...")


def extract_to_dir(filename, dirpath='.'):
    # Does not create folder twice with the same name
    name, ext = os.path.splitext(filename)
    # if os.path.basename(name) == os.path.basename(dirpath):
    #     dirpath = '.'
    # Extract
    print(dirpath)
    print("Extracting...", end="")
    if tarfile.is_tarfile(filename):
        tarfile.open(filename, 'r').extractall(dirpath)
    elif zipfile.is_zipfile(filename):
        zipfile.ZipFile(filename, 'r').extractall(dirpath)
    elif ext == '.gz':
        if not os.path.exists(dirpath):
            os.mkdir(dirpath)
        shutil.move(filename, os.path.join(dirpath, os.path.basename(filename)))
        print(f" | NOTE: gzip files are not extracted, and moved to {dirpath}", end="")
    # Return the path where the file was extracted
    print(" | Done !")
    return os.path.abspath(dirpath)

def progress_bar(current_index, max_index, prefix=None, suffix=None, start_time=None):
    """Display a progress bar and duration.

    Args:
        current_index (int): current state index (or epoch number).
        max_index (int): maximal numbers of state.
        prefix (str, optional): prefix of the progress bar. The default is None.
        suffix (str, optional): suffix of the progress bar. The default is None.
        start_time (float, optional): starting time of the progress bar. If not None, it will display the time
            spent from the beginning to the current state. The default is None.

    Returns:
        None. Display the progress bar in the console.
    """
    # Add a prefix to the progress bar
    prefix = "" if prefix is None else str(prefix) + " "

    # Get the percentage
    percentage = current_index * 100 // max_index
    loading = "[" + "=" * (percentage // 2) + " " * (50 - percentage // 2) + "]"
    progress_display = "\r{0}{1:3d}% | {2}".format(prefix, percentage, loading)

    # Add a suffix to the progress bar
    progress_display += "" if suffix is None else " | " + str(suffix)

    # Add a timer
    if start_time is not None:
        time_min, time_sec = get_time(start_time, time.time())
        time_display = " | Time: {0}m {1}s".format(time_min, time_sec)
        progress_display += time_display

    # Print the progress bar
    # TODO: return a string instead
    print(progress_display, end="{}".format("" if current_index < max_index else " | Done !\n"))

def get_time(start_time, end_time):
    """Get ellapsed time in minutes and seconds.

    Args:
        start_time (float): strarting time
        end_time (float): ending time

    Returns:
        elapsed_mins (float): elapsed time in minutes
        elapsed_secs (float): elapsed time in seconds.
    """
    elapsed_time = end_time - start_time
    elapsed_mins = int(elapsed_time / 60)
    elapsed_secs = int(elapsed_time - (elapsed_mins * 60))

    return elapsed_mins, elapsed_secs



from norch.tensor import Tensor
from norch.utils import utils
import random

class Parameter(Tensor):
    """
    A parameter is a trainable tensor.
    """
    def __init__(self, shape):
        data = utils.generate_random_list(shape=shape)
        super().__init__(data, requires_grad=True)


import norch

class ToTensor:
    def __call__(self, x):
        return norch.Tensor(x)

class Reshape:
    def __init__(self, shape):
        self.shape = shape

    def __call__(self, x):
        return x.reshape(self.shape)
    
class Sequential:
    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, x):
        for transform in self.transforms:
            x = transform(x)
        return x

